"""PolarsSeriesItem.

This module defines the PolarsSeriesItem class,
which represents a polars Series item.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from .item import Item, ItemTypeError

if TYPE_CHECKING:
    import polars


class PolarsSeriesItem(Item):
    """
    A class to represent a polars Series item.

    This class encapsulates a polars Series along with its
    creation and update timestamps.
    """

    def __init__(
        self,
        series_json: str,
        created_at: Union[str, None] = None,
        updated_at: Union[str, None] = None,
        note: Union[str, None] = None,
    ):
        """
        Initialize a PolarsSeriesItem.

        Parameters
        ----------
        series_json : str
            The JSON representation of the series, without its index.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        note : Union[str, None]
            An optional note.
        """
        super().__init__(created_at, updated_at, note)

        self.series_json = series_json

    @property
    def series(self) -> polars.Series:
        """
        The polars Series from the persistence.

        Its content can differ from the original series because it has been serialized
        using polars' `to_json` function and not pickled, in order to be
        environment-independent.
        """
        import io

        import polars

        with io.StringIO(self.series_json) as series_stream:
            series = polars.read_json(series_stream).to_series(0)

            return series

    @classmethod
    def factory(cls, series: polars.Series, /, **kwargs) -> PolarsSeriesItem:
        """
        Create a new PolarsSeriesItem instance from a polars Series.

        Parameters
        ----------
        series : polars.Series
            The polars Series to store.

        Returns
        -------
        PolarsSeriesItem
            A new PolarsSeriesItem instance.
        """
        import polars

        if not isinstance(series, polars.Series):
            raise ItemTypeError(f"Type '{series.__class__}' is not supported.")

        return cls(series_json=series.to_frame().write_json(), **kwargs)

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        return super().as_serializable_dict() | {
            "value": self.series.to_list(),
            "media_type": "text/markdown",
        }



import time

import joblib
import numpy as np
from sklearn.base import clone, is_classifier
from sklearn.model_selection import check_cv
from sklearn.pipeline import Pipeline

from skore.externals._pandas_accessors import DirNamesMixin
from skore.externals._sklearn_compat import _safe_indexing
from skore.sklearn._base import _BaseReport
from skore.sklearn._estimator.report import EstimatorReport
from skore.sklearn.find_ml_task import _find_ml_task
from skore.utils._progress_bar import progress_decorator


def _generate_estimator_report(estimator, X, y, train_indices, test_indices):
    return EstimatorReport(
        estimator,
        fit=True,
        X_train=_safe_indexing(X, train_indices),
        y_train=_safe_indexing(y, train_indices),
        X_test=_safe_indexing(X, test_indices),
        y_test=_safe_indexing(y, test_indices),
    )


class CrossValidationReport(_BaseReport, DirNamesMixin):
    """Reporter for cross-validation results.

    Parameters
    ----------
    estimator : estimator object
        Estimator to make the cross-validation report from.

    X : {array-like, sparse matrix} of shape (n_samples, n_features)
        The data to fit. Can be for example a list, or an array.

    y : array-like of shape (n_samples,) or (n_samples, n_outputs), default=None
        The target variable to try to predict in the case of supervised learning.

    cv_splitter : int, cross-validation generator or an iterable, default=5
        Determines the cross-validation splitting strategy.
        Possible inputs for `cv_splitter` are:

        - int, to specify the number of folds in a `(Stratified)KFold`,
        - a scikit-learn :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For int/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used. These splitters are instantiated
        with `shuffle=False` so the splits will be the same across calls.

        Refer to scikit-learn's :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

    n_jobs : int, default=None
        Number of jobs to run in parallel. Training the estimator and computing
        the score are parallelized over the cross-validation splits.
        When accessing some methods of the `CrossValidationReport`, the `n_jobs`
        parameter is used to parallelize the computation.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.

    Attributes
    ----------
    estimator_ : estimator object
        The cloned or copied estimator.

    estimator_name_ : str
        The name of the estimator.

    estimator_reports_ : list of EstimatorReport
        The estimator reports for each split.

    See Also
    --------
    skore.sklearn.estimator.report.EstimatorReport :
        Report for a fitted estimator.

    Examples
    --------
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.linear_model import LogisticRegression
    >>> X, y = make_classification(random_state=42)
    >>> estimator = LogisticRegression()
    >>> from skore import CrossValidationReport
    >>> report = CrossValidationReport(estimator, X=X, y=y, cv_splitter=2)
    Processing cross-validation ...
    """

    _ACCESSOR_CONFIG = {
        "metrics": {"name": "metrics"},
    }

    def __init__(
        self,
        estimator,
        X,
        y=None,
        cv_splitter=None,
        n_jobs=None,
    ):
        # used to know if a parent launch a progress bar manager
        self._parent_progress = None

        self._estimator = clone(estimator)

        # private storage to be able to invalidate the cache when the user alters
        # those attributes
        self._X = X
        self._y = y
        self._cv_splitter = check_cv(
            cv_splitter, y, classifier=is_classifier(estimator)
        )
        self.n_jobs = n_jobs

        self.estimator_reports_ = self._fit_estimator_reports()

        self._rng = np.random.default_rng(time.time_ns())
        self._hash = self._rng.integers(
            low=np.iinfo(np.int64).min, high=np.iinfo(np.int64).max
        )
        self._cache = {}
        self._ml_task = _find_ml_task(
            y, estimator=self.estimator_reports_[0]._estimator
        )

    @progress_decorator(
        description=lambda self: (
            f"Processing cross-validation\nfor {self.estimator_name_}"
        )
    )
    def _fit_estimator_reports(self):
        """Fit the estimator reports.

        This function is created to be able to use the progress bar. It works well
        with the patch of `rich` in VS Code.

        Returns
        -------
        estimator_reports : list of EstimatorReport
            The estimator reports.
        """
        progress = self._progress_info["current_progress"]
        task = self._progress_info["current_task"]

        n_splits = self._cv_splitter.get_n_splits(self._X, self._y)
        progress.update(task, total=n_splits)

        parallel = joblib.Parallel(n_jobs=self.n_jobs, return_as="generator")
        # do not split the data to take advantage of the memory mapping
        generator = parallel(
            joblib.delayed(_generate_estimator_report)(
                clone(self._estimator),
                self._X,
                self._y,
                train_indices,
                test_indices,
            )
            for train_indices, test_indices in self._cv_splitter.split(self._X, self._y)
        )

        estimator_reports = []
        for report in generator:
            estimator_reports.append(report)
            progress.update(task, advance=1, refresh=True)

        return estimator_reports

    def clear_cache(self):
        """Clear the cache.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import CrossValidationReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> reporter = CrossValidationReport(classifier, X=X, y=y, cv_splitter=2)
        Processing cross-validation ...
        >>> reporter.cache_predictions()
        Cross-validation predictions ...
        Caching predictions ...
        Caching predictions ...
        >>> reporter.clear_cache()
        >>> reporter._cache
        {}
        """
        for report in self.estimator_reports_:
            report.clear_cache()
        self._cache = {}

    @progress_decorator(description="Cross-validation predictions")
    def cache_predictions(self, response_methods="auto", n_jobs=None):
        """Cache the predictions for sub-estimators reports.

        Parameters
        ----------
        response_methods : {"auto", "predict", "predict_proba", "decision_function"},\
                default="auto
            The methods to use to compute the predictions.

        n_jobs : int, default=None
            The number of jobs to run in parallel. If `None`, we use the `n_jobs`
            parameter when initializing `CrossValidationReport`.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import CrossValidationReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> reporter = CrossValidationReport(classifier, X=X, y=y, cv_splitter=2)
        Processing cross-validation ...
        >>> reporter.cache_predictions()
        Cross-validation predictions ...
        Caching predictions ...
        Caching predictions ...
        >>> reporter._cache
        {...}
        """
        if n_jobs is None:
            n_jobs = self.n_jobs

        progress = self._progress_info["current_progress"]
        main_task = self._progress_info["current_task"]

        total_estimators = len(self.estimator_reports_)
        progress.update(main_task, total=total_estimators)

        for estimator_report in self.estimator_reports_:
            # Pass the progress manager to child tasks
            estimator_report._parent_progress = progress
            estimator_report.cache_predictions(
                response_methods=response_methods, n_jobs=n_jobs
            )
            progress.update(main_task, advance=1, refresh=True)

    @property
    def estimator_(self):
        return self._estimator

    @estimator_.setter
    def estimator_(self, value):
        raise AttributeError(
            "The estimator attribute is immutable. "
            f"Call the constructor of {self.__class__.__name__} to create a new report."
        )

    @property
    def estimator_name_(self):
        if isinstance(self._estimator, Pipeline):
            name = self._estimator[-1].__class__.__name__
        else:
            name = self._estimator.__class__.__name__
        return name

    @property
    def X(self):
        return self._X

    @X.setter
    def X(self, value):
        raise AttributeError(
            "The X attribute is immutable. "
            f"Call the constructor of {self.__class__.__name__} to create a new report."
        )

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        raise AttributeError(
            "The y attribute is immutable. "
            f"Call the constructor of {self.__class__.__name__} to create a new report."
        )

    ####################################################################################
    # Methods related to the help and repr
    ####################################################################################

    def _get_help_panel_title(self):
        return (
            f"[bold cyan]Tools to diagnose estimator "
            f"{self.estimator_name_}[/bold cyan]"
        )

    def _get_help_legend(self):
        return (
            "[cyan](↗︎)[/cyan] higher is better [orange1](↘︎)[/orange1] lower is better"
        )

    def __repr__(self):
        """Return a string representation using rich."""
        return self._rich_repr(
            class_name="skore.CrossValidationReport", help_method_name="reporter.help()"
        )



"""PlotlyFigureItem.

This module defines the PlotlyFigureItem class, used to persist Plotly figures.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .item import Item, ItemTypeError
from .media_item import lazy_is_instance

if TYPE_CHECKING:
    import plotly.basedatatypes


class PlotlyFigureItem(Item):
    """A class used to persist a Plotly figure."""

    def __init__(
        self,
        figure_str: str,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        note: Optional[str] = None,
    ):
        """
        Initialize a PlotlyFigureItem.

        Parameters
        ----------
        figure_str : bytes
            The JSON str of the Plotly figure.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        note : str, optional
            A note.
        """
        super().__init__(created_at, updated_at, note)

        self.figure_str = figure_str

    @classmethod
    def factory(
        cls,
        figure: plotly.basedatatypes.BaseFigure,
        /,
        **kwargs,
    ) -> PlotlyFigureItem:
        """
        Create a new PlotlyFigureItem instance from a Plotly figure.

        Parameters
        ----------
        figure : plotly.basedatatypes.BaseFigure
            The Plotly figure to store.

        Returns
        -------
        PlotlyFigureItem
            A new PlotlyFigureItem instance.
        """
        if not lazy_is_instance(figure, "plotly.basedatatypes.BaseFigure"):
            raise ItemTypeError(f"Type '{figure.__class__}' is not supported.")

        import plotly.io

        return cls(plotly.io.to_json(figure, engine="json"), **kwargs)

    @property
    def figure(self) -> plotly.basedatatypes.BaseFigure:
        """The figure from the persistence."""
        import plotly.io

        return plotly.io.from_json(self.figure_str)

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        import base64

        figure_bytes = self.figure_str.encode("utf-8")
        figure_b64_str = base64.b64encode(figure_bytes).decode()

        return super().as_serializable_dict() | {
            "media_type": "application/vnd.plotly.v1+json;base64",
            "value": figure_b64_str,
        }



"""Configure logging and global settings."""

import logging

from rich.console import Console
from rich.theme import Theme

from skore.project import Project, open
from skore.sklearn import (
    CrossValidationReport,
    CrossValidationReporter,
    EstimatorReport,
    train_test_split,
)
from skore.utils._patch import setup_jupyter_display
from skore.utils._show_versions import show_versions

__all__ = [
    "CrossValidationReporter",
    "CrossValidationReport",
    "EstimatorReport",
    "open",
    "Project",
    "show_versions",
    "train_test_split",
]

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())  # Default to no output
logger.setLevel(logging.INFO)

# Configure jupyter display for VS Code compatibility
setup_jupyter_display()

skore_console_theme = Theme(
    {
        "repr.str": "cyan",
        "rule.line": "orange1",
        "repr.url": "orange1",
    }
)

console = Console(theme=skore_console_theme, width=88)



"""PillowImageItem.

This module defines the PillowImageItem class, used to persist Pillow images.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .item import Item, ItemTypeError
from .media_item import lazy_is_instance

if TYPE_CHECKING:
    import PIL.Image


class PillowImageItem(Item):
    """A class used to persist a Pillow image."""

    def __init__(
        self,
        image_bytes: bytes,
        image_mode: str,
        image_size: tuple[int],
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        note: Optional[str] = None,
    ):
        """
        Initialize a PillowImageItem.

        Parameters
        ----------
        image_bytes : bytes
            The raw bytes of the Pillow image.
        image_mode : str
            The image mode.
        image_size : tuple[int]
            The image size.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        note : str, optional
            A note.
        """
        super().__init__(created_at, updated_at, note)

        self.image_bytes = image_bytes
        self.image_mode = image_mode
        self.image_size = image_size

    @classmethod
    def factory(cls, image: PIL.Image.Image, /, **kwargs) -> PillowImageItem:
        """
        Create a new PillowImageItem instance from a Pillow image.

        Parameters
        ----------
        image : PIL.Image.Image
            The Pillow image to store.

        Returns
        -------
        PillowImageItem
            A new PillowImageItem instance.
        """
        if not lazy_is_instance(image, "PIL.Image.Image"):
            raise ItemTypeError(f"Type '{image.__class__}' is not supported.")

        return cls(
            image_bytes=image.tobytes(),
            image_mode=image.mode,
            image_size=image.size,
            **kwargs,
        )

    @property
    def image(self) -> PIL.Image.Image:
        """The image from the persistence."""
        import PIL.Image

        return PIL.Image.frombytes(
            mode=self.image_mode,
            size=self.image_size,
            data=self.image_bytes,
        )

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        import base64
        import io

        with io.BytesIO() as stream:
            self.image.save(stream, format="png")

            png_bytes = stream.getvalue()
            png_b64_str = base64.b64encode(png_bytes).decode()

        return super().as_serializable_dict() | {
            "media_type": "image/png;base64",
            "value": png_b64_str,
        }



"""Implement the "launch" command."""

import webbrowser
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Union

import uvicorn
from fastapi import FastAPI

from skore.cli import logger
from skore.project import open
from skore.ui.app import create_app
from skore.utils._logger import logger_context


def __launch(
    project_name: Union[str, Path],
    port: int,
    open_browser: bool,
    verbose: bool = False,
):
    """Launch the UI to visualize a project.

    Parameters
    ----------
    project_name : Path-like
        Name of the project to be created, or a relative or absolute path.
    port : int
        Port at which to bind the UI server.
    open_browser: bool
        Whether to automatically open a browser tab showing the UI.
    verbose: bool
        Whether to display info logs to the user.
    """
    from skore import console  # avoid circular import

    with logger_context(logger, verbose):
        project = open(project_name)

        @asynccontextmanager
        async def lifespan(app: FastAPI):
            if open_browser:
                webbrowser.open(f"http://localhost:{port}")
            yield

        app = create_app(project=project, lifespan=lifespan)

        try:
            # TODO: check port is free
            console.rule("[bold cyan]skore-UI[/bold cyan]")
            console.print(
                f"Running skore UI from '{project_name}' at URL http://localhost:{port}"
            )
            uvicorn.run(app, port=port, log_level="error")
        except KeyboardInterrupt:
            console.print("Closing skore UI")



"""Stratify warning.

This warning is shown when `stratify` is set.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from skore.sklearn.train_test_split.warning.train_test_split_warning import (
    TrainTestSplitWarning,
)

if TYPE_CHECKING:
    ArrayLike = Any


class StratifyWarning(TrainTestSplitWarning):
    """Check whether `stratify` is set."""

    MSG = (
        "We recommend against using `stratify` if you have a problem where your events "
        "occur over time (even if they are independent). Using `stratify` will result "
        "in inflated model performance evaluation because the natural drift of the "
        "target will not be taken into account."
    )

    @staticmethod
    def check(stratify: Optional[ArrayLike], **kwargs) -> Union[str, None]:
        """Check whether stratify is set.

        Parameters
        ----------
        stratify : array-like or None
            An 1-dimensional target array to be used for stratification.

        Returns
        -------
        warning
            None if the check passed, otherwise the warning message.
        """
        if stratify:
            return StratifyWarning.MSG



"""PandasSeriesItem.

This module defines the PandasSeriesItem class,
which represents a pandas Series item.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from .item import Item, ItemTypeError

if TYPE_CHECKING:
    import pandas


class PandasSeriesItem(Item):
    """
    A class to represent a pandas Series item.

    This class encapsulates a pandas Series along with its
    creation and update timestamps.
    """

    ORIENT = "split"

    def __init__(
        self,
        index_json: str,
        series_json: str,
        created_at: Union[str, None] = None,
        updated_at: Union[str, None] = None,
        note: Union[str, None] = None,
    ):
        """
        Initialize a PandasSeriesItem.

        Parameters
        ----------
        index_json : str
            The JSON representation of the series's index.
        series_json : str
            The JSON representation of the series, without its index.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        note : Union[str, None]
            An optional note.
        """
        super().__init__(created_at, updated_at, note)

        self.index_json = index_json
        self.series_json = series_json

    @property
    def series(self) -> pandas.Series:
        """
        The pandas Series from the persistence.

        Its content can differ from the original series because it has been serialized
        using pandas' `to_json` function and not pickled, in order to be
        environment-independent.
        """
        import io

        import pandas

        with (
            io.StringIO(self.index_json) as index_stream,
            io.StringIO(self.series_json) as series_stream,
        ):
            index = pandas.read_json(index_stream, orient=self.ORIENT, dtype=False)
            index = index.set_index(list(index.columns))
            series = pandas.read_json(
                series_stream,
                typ="series",
                orient=self.ORIENT,
                dtype=False,
            )
            series.index = index.index

            return series

    @classmethod
    def factory(cls, series: pandas.Series, /, **kwargs) -> PandasSeriesItem:
        """
        Create a new PandasSeriesItem instance from a pandas Series.

        Parameters
        ----------
        series : pd.Series
            The pandas Series to store.

        Returns
        -------
        PandasSeriesItem
            A new PandasSeriesItem instance.
        """
        import pandas

        if not isinstance(series, pandas.Series):
            raise ItemTypeError(f"Type '{series.__class__}' is not supported.")

        # One native method is available to serialize series with multi-index,
        # while keeping the index names:
        #
        # Using table orientation with JSON serializer:
        #    ```python
        #    json = series.to_json(orient="table")
        #    series = pandas.read_json(json, typ="series", orient="table", dtype=False)
        #    ```
        #
        #    This method fails when an index name is an integer.
        #
        # None of those methods being compatible, we store indexes separately.

        index = series.index.to_frame(index=False)
        series = series.reset_index(drop=True)

        return cls(
            index_json=index.to_json(orient=PandasSeriesItem.ORIENT),
            series_json=series.to_json(orient=PandasSeriesItem.ORIENT),
            **kwargs,
        )

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        return super().as_serializable_dict() | {
            "value": self.series.fillna("NaN").to_list(),
            "media_type": "text/markdown",
        }



import numbers

import matplotlib.pyplot as plt
import numpy as np
from sklearn.utils.validation import check_random_state

from skore.externals._sklearn_compat import _safe_indexing
from skore.sklearn._plot.utils import (
    HelpDisplayMixin,
    _despine_matplotlib_axis,
    _validate_style_kwargs,
    sample_mpl_colormap,
)


class PredictionErrorDisplay(HelpDisplayMixin):
    """Visualization of the prediction error of a regression model.

    This tool can display "residuals vs predicted" or "actual vs predicted"
    using scatter plots to qualitatively assess the behavior of a regressor,
    preferably on held-out data points.

    An instance of this class is should created by
    `EstimatorReport.metrics.plot.prediction_error()`.
    You should not create an instance of this class directly.

    Parameters
    ----------
    y_true : list of ndarray of shape (n_samples,)
        True values.

    y_pred : list of ndarray of shape (n_samples,)
        Prediction values.

    estimator_name : str
        Name of the estimator.

    data_source : {"train", "test", "X_y"}, default=None
        The data source used to display the prediction error.

    Attributes
    ----------
    line_ : matplotlib Artist
        Optimal line representing `y_true == y_pred`. Therefore, it is a
        diagonal line for `kind="predictions"` and a horizontal line for
        `kind="residuals"`.

    errors_lines_ : matplotlib Artist or None
        Residual lines. If `with_errors=False`, then it is set to `None`.

    scatter_ : matplotlib Artist
        Scatter data points.

    ax_ : matplotlib Axes
        Axes with the different matplotlib axis.

    figure_ : matplotlib Figure
        Figure containing the scatter and lines.

    Examples
    --------
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.linear_model import Ridge
    >>> from sklearn.model_selection import train_test_split
    >>> from skore import EstimatorReport
    >>> X_train, X_test, y_train, y_test = train_test_split(
    ...     *load_diabetes(return_X_y=True), random_state=0
    ... )
    >>> classifier = Ridge()
    >>> reporter = EstimatorReport(
    ...     classifier,
    ...     X_train=X_train,
    ...     y_train=y_train,
    ...     X_test=X_test,
    ...     y_test=y_test,
    ... )
    >>> display = reporter.metrics.plot.prediction_error()
    >>> display.plot(kind="actual_vs_predicted")
    """

    def __init__(self, *, y_true, y_pred, estimator_name, data_source=None):
        self.y_true = y_true
        self.y_pred = y_pred
        self.estimator_name = estimator_name
        self.data_source = data_source

    def plot(
        self,
        ax=None,
        *,
        estimator_name=None,
        kind="residual_vs_predicted",
        scatter_kwargs=None,
        line_kwargs=None,
        despine=True,
    ):
        """Plot visualization.

        Extra keyword arguments will be passed to matplotlib's ``plot``.

        Parameters
        ----------
        ax : matplotlib axes, default=None
            Axes object to plot on. If `None`, a new figure and axes is
            created.

        estimator_name : str
            Name of the estimator used to plot the prediction error. If `None`,
            we used the inferred name from the estimator.

        kind : {"actual_vs_predicted", "residual_vs_predicted"}, \
                default="residual_vs_predicted"
            The type of plot to draw:

            - "actual_vs_predicted" draws the observed values (y-axis) vs.
              the predicted values (x-axis).
            - "residual_vs_predicted" draws the residuals, i.e. difference
              between observed and predicted values, (y-axis) vs. the predicted
              values (x-axis).

        scatter_kwargs : dict, default=None
            Dictionary with keywords passed to the `matplotlib.pyplot.scatter`
            call.

        line_kwargs : dict, default=None
            Dictionary with keyword passed to the `matplotlib.pyplot.plot`
            call to draw the optimal line.

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Returns
        -------
        display : PredictionErrorDisplay
            Object that stores computed values.

        Examples
        --------
        >>> from sklearn.datasets import load_diabetes
        >>> from sklearn.linear_model import Ridge
        >>> from sklearn.model_selection import train_test_split
        >>> from skore import EstimatorReport
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     *load_diabetes(return_X_y=True), random_state=0
        ... )
        >>> classifier = Ridge()
        >>> reporter = EstimatorReport(
        ...     classifier,
        ...     X_train=X_train,
        ...     y_train=y_train,
        ...     X_test=X_test,
        ...     y_test=y_test,
        ... )
        >>> display = reporter.metrics.plot.prediction_error()
        >>> display.plot(kind="actual_vs_predicted")
        """
        expected_kind = ("actual_vs_predicted", "residual_vs_predicted")
        if kind not in expected_kind:
            raise ValueError(
                f"`kind` must be one of {', '.join(expected_kind)}. "
                f"Got {kind!r} instead."
            )
        if kind == "actual_vs_predicted":
            xlabel, ylabel = "Predicted values", "Actual values"
        else:  # kind == "residual_vs_predicted"
            xlabel, ylabel = "Predicted values", "Residuals (actual - predicted)"

        if scatter_kwargs is None:
            scatter_kwargs = {}
        if line_kwargs is None:
            line_kwargs = {}

        if estimator_name is None:
            estimator_name = self.estimator_name

        if ax is None:
            _, ax = plt.subplots()

        x_range_perfect_pred = [np.inf, -np.inf]
        y_range_perfect_pred = [np.inf, -np.inf]
        for y_true, y_pred in zip(self.y_true, self.y_pred):
            if kind == "actual_vs_predicted":
                min_value = min(y_pred.min(), y_true.min())
                max_value = max(y_pred.max(), y_true.max())
                x_range_perfect_pred[0] = min(x_range_perfect_pred[0], min_value)
                x_range_perfect_pred[1] = max(x_range_perfect_pred[1], max_value)
                y_range_perfect_pred[0] = min(y_range_perfect_pred[0], min_value)
                y_range_perfect_pred[1] = max(y_range_perfect_pred[1], max_value)
            else:
                residuals = y_true - y_pred
                x_range_perfect_pred[0] = min(x_range_perfect_pred[0], y_pred.min())
                x_range_perfect_pred[1] = max(x_range_perfect_pred[1], y_pred.max())
                y_range_perfect_pred[0] = min(y_range_perfect_pred[0], residuals.min())
                y_range_perfect_pred[1] = max(y_range_perfect_pred[1], residuals.max())

        colors_markers = sample_mpl_colormap(
            plt.cm.tab10, len(self.y_true) if len(self.y_true) > 10 else 10
        )
        for split_idx in range(len(self.y_true)):
            y_true, y_pred = self.y_true[split_idx], self.y_pred[split_idx]

            default_scatter_kwargs = {
                "color": colors_markers[split_idx],
                "alpha": 0.3,
                "s": 10,
            }
            prediction_error_scatter_kwargs = _validate_style_kwargs(
                default_scatter_kwargs, scatter_kwargs
            )

            if self.data_source in ("train", "test"):
                scatter_label = f"{self.data_source.title()} set"
            else:  # data_source == "X_y"
                scatter_label = "Data set"

            if len(self.y_true) > 1:  # cross-validation
                scatter_label += f" - split #{split_idx + 1}"

            if kind == "actual_vs_predicted":
                self.scatter_ = ax.scatter(
                    y_pred,
                    y_true,
                    label=scatter_label,
                    **prediction_error_scatter_kwargs,
                )

            else:  # kind == "residual_vs_predicted"
                residuals = y_true - y_pred
                self.scatter_ = ax.scatter(
                    y_pred,
                    residuals,
                    label=scatter_label,
                    **prediction_error_scatter_kwargs,
                )

        default_line_kwargs = {
            "color": "black",
            "alpha": 0.7,
            "linestyle": "--",
            "label": "Perfect predictions",
        }
        perfect_line_kwargs = _validate_style_kwargs(default_line_kwargs, line_kwargs)

        if kind == "actual_vs_predicted":
            self.line_ = ax.plot(
                x_range_perfect_pred, y_range_perfect_pred, **perfect_line_kwargs
            )[0]
            ax.set(
                aspect="equal",
                xlim=x_range_perfect_pred,
                ylim=y_range_perfect_pred,
                xticks=np.linspace(
                    x_range_perfect_pred[0], x_range_perfect_pred[1], num=5
                ),
                yticks=np.linspace(
                    y_range_perfect_pred[0], y_range_perfect_pred[1], num=5
                ),
            )
        else:
            self.line_ = ax.plot(x_range_perfect_pred, [0, 0], **perfect_line_kwargs)[0]
            ax.set(
                aspect="equal",
                xlim=x_range_perfect_pred,
                ylim=y_range_perfect_pred,
                xticks=np.linspace(
                    x_range_perfect_pred[0], x_range_perfect_pred[1], num=5
                ),
                yticks=np.linspace(
                    y_range_perfect_pred[0], y_range_perfect_pred[1], num=5
                ),
            )

        ax.set(xlabel=xlabel, ylabel=ylabel)
        ax.legend(title=estimator_name)

        self.ax_ = ax
        self.figure_ = ax.figure

        if despine:
            x_range = self.ax_.get_xlim()
            y_range = self.ax_.get_ylim()
            _despine_matplotlib_axis(self.ax_, x_range=x_range, y_range=y_range)

    @classmethod
    def _from_predictions(
        cls,
        y_true,
        y_pred,
        *,
        estimator,  # currently only for consistency with other plots
        estimator_name,
        ml_task,  # FIXME: to be used when having single-output vs. multi-output
        data_source=None,
        kind="residual_vs_predicted",
        subsample=1_000,
        random_state=None,
        ax=None,
        scatter_kwargs=None,
        line_kwargs=None,
        despine=True,
    ):
        """Plot the prediction error given the true and predicted targets.

        Parameters
        ----------
        y_true : list of array-like of shape (n_samples,)
            True target values.

        y_pred : list of array-like of shape (n_samples,)
            Predicted target values.

        estimator : estimator instance
            The estimator from which `y_pred` is obtained.

        estimator_name : str,
            The name of the estimator.

        ml_task : {"binary-classification", "multiclass-classification"}
            The machine learning task.

        data_source : {"train", "test", "X_y"}, default=None
            The data source used to compute the ROC curve.

        kind : {"actual_vs_predicted", "residual_vs_predicted"}, \
                default="residual_vs_predicted"
            The type of plot to draw:

            - "actual_vs_predicted" draws the observed values (y-axis) vs.
              the predicted values (x-axis).
            - "residual_vs_predicted" draws the residuals, i.e. difference
              between observed and predicted values, (y-axis) vs. the predicted
              values (x-axis).

        subsample : float, int or None, default=1_000
            Sampling the samples to be shown on the scatter plot. If `float`,
            it should be between 0 and 1 and represents the proportion of the
            original dataset. If `int`, it represents the number of samples
            display on the scatter plot. If `None`, no subsampling will be
            applied. by default, 1000 samples or less will be displayed.

        random_state : int or RandomState, default=None
            Controls the randomness when `subsample` is not `None`.
            See :term:`Glossary <random_state>` for details.

        ax : matplotlib axes, default=None
            Axes object to plot on. If `None`, a new figure and axes is
            created.

        scatter_kwargs : dict, default=None
            Dictionary with keywords passed to the `matplotlib.pyplot.scatter`
            call.

        line_kwargs : dict, default=None
            Dictionary with keyword passed to the `matplotlib.pyplot.plot`
            call to draw the optimal line.

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Returns
        -------
        display : PredictionErrorDisplay
            Object that stores the computed values.
        """
        random_state = check_random_state(random_state)
        if isinstance(subsample, numbers.Integral):
            if subsample <= 0:
                raise ValueError(
                    f"When an integer, subsample={subsample} should be positive."
                )
        elif isinstance(subsample, numbers.Real) and (subsample <= 0 or subsample >= 1):
            raise ValueError(
                f"When a floating-point, subsample={subsample} should be in the "
                "(0, 1) range."
            )

        y_true_display, y_pred_display = [], []
        for y_true_i, y_pred_i in zip(y_true, y_pred):
            n_samples = len(y_true_i)
            if subsample is None:
                subsample_ = n_samples
            elif isinstance(subsample, numbers.Integral):
                subsample_ = subsample
            else:  # subsample is a float
                subsample_ = int(n_samples * subsample)

            # normalize subsample based on the number of splits
            subsample_ = int(subsample_ / len(y_true))
            if subsample_ < n_samples:
                indices = random_state.choice(np.arange(n_samples), size=subsample_)
                y_true_display.append(_safe_indexing(y_true_i, indices, axis=0))
                y_pred_display.append(_safe_indexing(y_pred_i, indices, axis=0))
            else:
                y_true_display.append(y_true_i)
                y_pred_display.append(y_pred_i)

        viz = cls(
            y_true=y_true_display,
            y_pred=y_pred_display,
            estimator_name=estimator_name,
            data_source=data_source,
        )

        viz.plot(
            ax=ax,
            estimator_name=estimator_name,
            kind=kind,
            scatter_kwargs=scatter_kwargs,
            line_kwargs=line_kwargs,
            despine=despine,
        )

        return viz



import copy
import time
import warnings
from itertools import product

import joblib
import numpy as np
from sklearn.base import clone
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from skore.externals._pandas_accessors import DirNamesMixin
from skore.externals._sklearn_compat import is_clusterer
from skore.sklearn._base import _BaseReport, _get_cached_response_values
from skore.sklearn.find_ml_task import _find_ml_task
from skore.utils._progress_bar import progress_decorator


class EstimatorReport(_BaseReport, DirNamesMixin):
    """Report for a fitted estimator.

    This class provides a set of tools to quickly validate and inspect a scikit-learn
    compatible estimator.

    Parameters
    ----------
    estimator : estimator object
        Estimator to make the report from. When the estimator is not fitted,
        it is deep-copied to avoid side-effects. If it is fitted, it is cloned instead.

    fit : {"auto", True, False}, default="auto"
        Whether to fit the estimator on the training data. If "auto", the estimator
        is fitted only if the training data is provided.

    X_train : {array-like, sparse matrix} of shape (n_samples, n_features) or \
            None
        Training data.

    y_train : array-like of shape (n_samples,) or (n_samples, n_outputs) or None
        Training target.

    X_test : {array-like, sparse matrix} of shape (n_samples, n_features) or None
        Testing data. It should have the same structure as the training data.

    y_test : array-like of shape (n_samples,) or (n_samples, n_outputs) or None
        Testing target.

    Attributes
    ----------
    estimator_ : estimator object
        The cloned or copied estimator.

    estimator_name_ : str
        The name of the estimator.

    See Also
    --------
    skore.sklearn.cross_validation.report.CrossValidationReport
        Report of cross-validation results.

    Examples
    --------
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.linear_model import LogisticRegression
    >>> X, y = make_classification(random_state=42)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)
    >>> estimator = LogisticRegression().fit(X_train, y_train)
    >>> from skore import EstimatorReport
    >>> report = EstimatorReport(estimator, X_test=X_test, y_test=y_test)
    """

    _ACCESSOR_CONFIG = {
        "metrics": {"name": "metrics"},
    }

    @staticmethod
    def _fit_estimator(estimator, X_train, y_train):
        if X_train is None or (y_train is None and not is_clusterer(estimator)):
            raise ValueError(
                "The training data is required to fit the estimator. "
                "Please provide both X_train and y_train."
            )
        return clone(estimator).fit(X_train, y_train)

    @classmethod
    def _copy_estimator(cls, estimator):
        try:
            return copy.deepcopy(estimator)
        except Exception as e:
            warnings.warn(
                "Deepcopy failed; using estimator as-is. "
                "Be aware that modifying the estimator outside of "
                f"{cls.__name__} will modify the internal estimator. "
                "Consider using a FrozenEstimator from scikit-learn to prevent this. "
                f"Original error: {e}",
                stacklevel=1,
            )
        return estimator

    def __init__(
        self,
        estimator,
        *,
        fit="auto",
        X_train=None,
        y_train=None,
        X_test=None,
        y_test=None,
    ):
        # used to know if a parent launch a progress bar manager
        self._parent_progress = None

        if fit == "auto":
            try:
                check_is_fitted(estimator)
                self._estimator = self._copy_estimator(estimator)
            except NotFittedError:
                self._estimator = self._fit_estimator(estimator, X_train, y_train)
        elif fit is True:
            self._estimator = self._fit_estimator(estimator, X_train, y_train)
        else:  # fit is False
            self._estimator = self._copy_estimator(estimator)

        # private storage to be able to invalidate the cache when the user alters
        # those attributes
        self._X_train = X_train
        self._y_train = y_train
        self._X_test = X_test
        self._y_test = y_test

        self._initialize_state()

    def _initialize_state(self):
        """Initialize/reset the random number generator, hash, and cache."""
        self._rng = np.random.default_rng(time.time_ns())
        self._hash = self._rng.integers(
            low=np.iinfo(np.int64).min, high=np.iinfo(np.int64).max
        )
        self._cache = {}
        self._ml_task = _find_ml_task(self._y_test, estimator=self._estimator)

    # NOTE:
    # For the moment, we do not allow to alter the estimator and the training data.
    # For the validation set, we allow it and we invalidate the cache.

    def clear_cache(self):
        """Clear the cache.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from sklearn.model_selection import train_test_split
        >>> from skore import EstimatorReport
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     *load_breast_cancer(return_X_y=True), random_state=0
        ... )
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> reporter = EstimatorReport(
        ...     classifier,
        ...     X_train=X_train,
        ...     y_train=y_train,
        ...     X_test=X_test,
        ...     y_test=y_test,
        ... )
        >>> reporter.cache_predictions()
        Caching predictions ...
        >>> reporter.clear_cache()
        >>> reporter._cache
        {}
        """
        self._cache = {}

    @progress_decorator(description="Caching predictions")
    def cache_predictions(self, response_methods="auto", n_jobs=None):
        """Cache estimator's predictions.

        Parameters
        ----------
        response_methods : "auto" or list of str, default="auto"
            The response methods to precompute. If "auto", the response methods are
            inferred from the ml task: for classification we compute the response of
            the `predict_proba`, `decision_function` and `predict` methods; for
            regression we compute the response of the `predict` method.

        n_jobs : int or None, default=None
            The number of jobs to run in parallel. None means 1 unless in a
            joblib.parallel_backend context. -1 means using all processors.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from sklearn.model_selection import train_test_split
        >>> from skore import EstimatorReport
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     *load_breast_cancer(return_X_y=True), random_state=0
        ... )
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> reporter = EstimatorReport(
        ...     classifier,
        ...     X_train=X_train,
        ...     y_train=y_train,
        ...     X_test=X_test,
        ...     y_test=y_test,
        ... )
        >>> reporter.cache_predictions()
        Caching predictions ...
        >>> reporter._cache
        {...}
        """
        if self._ml_task in ("binary-classification", "multiclass-classification"):
            if response_methods == "auto":
                response_methods = ["predict"]
                if hasattr(self._estimator, "predict_proba"):
                    response_methods += ["predict_proba"]
                if hasattr(self._estimator, "decision_function"):
                    response_methods += ["decision_function"]
            pos_labels = self._estimator.classes_.tolist() + [None]
        else:
            if response_methods == "auto":
                response_methods = ["predict"]
            pos_labels = [None]

        data_sources = [("test", self._X_test)]
        if self._X_train is not None:
            data_sources += [("train", self._X_train)]

        parallel = joblib.Parallel(
            n_jobs=n_jobs, return_as="generator", require="sharedmem"
        )
        generator = parallel(
            joblib.delayed(_get_cached_response_values)(
                cache=self._cache,
                estimator_hash=self._hash,
                estimator=self._estimator,
                X=X,
                response_method=response_method,
                pos_label=pos_label,
                data_source=data_source,
            )
            for response_method, pos_label, (data_source, X) in product(
                response_methods, pos_labels, data_sources
            )
        )
        # trigger the computation
        progress = self._progress_info["current_progress"]
        task = self._progress_info["current_task"]
        total_iterations = len(response_methods) * len(pos_labels) * len(data_sources)
        progress.update(task, total=total_iterations)
        for _ in generator:
            progress.update(task, advance=1, refresh=True)

    @property
    def estimator_(self):
        return self._estimator

    @estimator_.setter
    def estimator_(self, value):
        raise AttributeError(
            "The estimator attribute is immutable. "
            f"Call the constructor of {self.__class__.__name__} to create a new report."
        )

    @property
    def X_train(self):
        return self._X_train

    @X_train.setter
    def X_train(self, value):
        raise AttributeError(
            "The X_train attribute is immutable. "
            f"Call the constructor of {self.__class__.__name__} to create a new report."
        )

    @property
    def y_train(self):
        return self._y_train

    @y_train.setter
    def y_train(self, value):
        raise AttributeError(
            "The y_train attribute is immutable. "
            f"Call the constructor of {self.__class__.__name__} to create a new report."
        )

    @property
    def X_test(self):
        return self._X_test

    @X_test.setter
    def X_test(self, value):
        self._X_test = value
        self._initialize_state()

    @property
    def y_test(self):
        return self._y_test

    @y_test.setter
    def y_test(self, value):
        self._y_test = value
        self._initialize_state()

    @property
    def estimator_name_(self):
        if isinstance(self._estimator, Pipeline):
            name = self._estimator[-1].__class__.__name__
        else:
            name = self._estimator.__class__.__name__
        return name

    ####################################################################################
    # Methods related to the help and repr
    ####################################################################################

    def _get_help_panel_title(self):
        return (
            f"[bold cyan]Tools to diagnose estimator "
            f"{self.estimator_name_}[/bold cyan]"
        )

    def _get_help_legend(self):
        return (
            "[cyan](↗︎)[/cyan] higher is better [orange1](↘︎)[/orange1] lower is better"
        )

    def __repr__(self):
        """Return a string representation using rich."""
        return self._rich_repr(
            class_name="skore.EstimatorReport", help_method_name="reporter.help()"
        )



"""Plot cross-validation timing results."""

from __future__ import annotations

from typing import TYPE_CHECKING

from numpy import linspace

if TYPE_CHECKING:
    import plotly.graph_objects


def plot_cross_validation_timing(cv_results: dict) -> plotly.graph_objects.Figure:
    """Plot the timing results of a cross-validation run.

    Parameters
    ----------
    cv_results : dict
        The output of scikit-learn's cross_validate function.

    Returns
    -------
    plotly.graph_objects.Figure
        A plot of the time-related cross-validation results.
    """
    from datetime import timedelta

    import pandas
    import plotly
    import plotly.graph_objects as go

    _cv_results = cv_results.copy()

    # Remove irrelevant keys
    to_remove = [key for key in _cv_results if key not in ["fit_time", "score_time"]]
    for key in to_remove:
        _cv_results.pop(key, None)

    df = pandas.DataFrame(_cv_results)

    dict_labels = {
        "fit_time": "fit_time (seconds)",
        "score_time": "score_time (seconds)",
    }

    fig = go.Figure()

    for col_i, col_name in enumerate(df.columns):
        metric_name = dict_labels.get(col_name, col_name)
        bar_color = plotly.colors.qualitative.Plotly[
            col_i % len(plotly.colors.qualitative.Plotly)
        ]
        bar_x = linspace(min(df.index) - 0.5, max(df.index) + 0.5, num=10)

        common_kwargs = dict(
            visible=True if col_i == 0 else "legendonly",
            legendgroup=f"group{col_i}",
        )

        # Calculate statistics
        avg_value = df[col_name].mean()
        std_value = df[col_name].std()

        # Add all traces at once
        fig.add_traces(
            [
                # Bar trace
                go.Bar(
                    x=df.index,
                    y=df[col_name].values,
                    name=metric_name,
                    marker_color=bar_color,
                    showlegend=True,
                    hovertemplate=(
                        "%{customdata}" f"<extra>{col_name} (timedelta)</extra>"
                    ),
                    customdata=[str(timedelta(seconds=x)) for x in df[col_name].values],
                    **common_kwargs,
                ),
                # Mean line
                go.Scatter(
                    x=bar_x,
                    y=[avg_value] * 10,
                    name=f"Average {metric_name}",
                    line=dict(dash="dash", color=bar_color),
                    showlegend=False,
                    mode="lines",
                    hovertemplate="%{customdata}",
                    customdata=[str(timedelta(seconds=avg_value))] * 10,
                    **common_kwargs,
                ),
                # +1 std line
                go.Scatter(
                    x=bar_x,
                    y=[avg_value + std_value] * 10,
                    name=f"Average + 1 std. dev. {metric_name}",
                    line=dict(dash="dot", color=bar_color),
                    showlegend=False,
                    mode="lines",
                    hovertemplate="%{customdata}",
                    customdata=[str(timedelta(seconds=avg_value + std_value))] * 10,
                    **common_kwargs,
                ),
                # -1 std line
                go.Scatter(
                    x=bar_x,
                    y=[avg_value - std_value] * 10,
                    name=f"Average - 1 std. dev. {metric_name}",
                    line=dict(dash="dot", color=bar_color),
                    showlegend=False,
                    mode="lines",
                    hovertemplate="%{customdata}",
                    customdata=[str(timedelta(seconds=avg_value + std_value))] * 10,
                    **common_kwargs,
                ),
            ]
        )

    fig.update_xaxes(tickmode="linear", dtick=1, title_text="Split index")
    fig.update_yaxes(title_text="Value")
    fig.update_layout(title_text="Time-related cross-validation results for each split")

    return fig



from typing import Any, Literal, Optional, Union

import numpy as np
from rich.panel import Panel
from rich.tree import Tree
from sklearn.base import BaseEstimator

class _HelpMixin:
    def _get_methods_for_help(self) -> list[tuple[str, Any]]: ...
    def _sort_methods_for_help(
        self, methods: list[tuple[str, Any]]
    ) -> list[tuple[str, Any]]: ...
    def _format_method_name(self, name: str) -> str: ...
    def _get_method_description(self, method: Any) -> str: ...
    def _create_help_panel(self) -> Panel: ...
    def _get_help_panel_title(self) -> str: ...
    def _create_help_tree(self) -> Tree: ...
    def help(self) -> None: ...
    def __repr__(self) -> str: ...

class _BaseAccessor(_HelpMixin):
    _parent: Any

    def __init__(self, parent: Any) -> None: ...
    def _get_help_panel_title(self) -> str: ...
    def _create_help_tree(self) -> Tree: ...
    def _get_X_y_and_data_source_hash(
        self,
        *,
        data_source: Literal["test", "train", "X_y"],
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
    ) -> tuple[Optional[np.ndarray], Optional[np.ndarray], Optional[str]]: ...

def _get_cached_response_values(
    *,
    cache: dict,
    estimator_hash: int,
    estimator: BaseEstimator,
    X: np.ndarray,
    response_method: Union[str, list[str]],
    pos_label: Optional[Union[str, int]] = None,
    data_source: Literal["test", "train", "X_y"] = "test",
    data_source_hash: Optional[str] = None,
) -> np.ndarray: ...

class _BaseReport(_HelpMixin):
    _ACCESSOR_CONFIG: dict[str, dict[str, str]]
    estimator_: BaseEstimator
    estimator_name_: str
    _X_test: Optional[np.ndarray]
    _y_test: Optional[np.ndarray]
    _X_train: Optional[np.ndarray]
    _y_train: Optional[np.ndarray]

    def _get_help_panel_title(self) -> str: ...
    def _get_help_legend(self) -> str: ...
    def _get_attributes_for_help(self) -> list[str]: ...
    def _create_help_tree(self) -> Tree: ...



from skore.externals._pandas_accessors import _register_accessor
from skore.sklearn._estimator.metrics_accessor import (
    _MetricsAccessor,
    _PlotMetricsAccessor,
)
from skore.sklearn._estimator.report import EstimatorReport

# add the metrics accessor to the estimator report
_register_accessor("metrics", EstimatorReport)(_MetricsAccessor)

# add the plot accessor to the metrics accessor
_register_accessor("plot", _MetricsAccessor)(_PlotMetricsAccessor)

__all__ = ["EstimatorReport"]



"""FastAPI factory used to create the API to interact with stores."""

import sys
from pathlib import Path
from typing import Optional

from fastapi import APIRouter, FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles
from starlette.types import Lifespan

from skore.project import Project, open
from skore.ui.dependencies import get_static_path
from skore.ui.project_routes import router as project_router


def create_app(
    project: Optional[Project] = None, lifespan: Optional[Lifespan] = None
) -> FastAPI:
    """FastAPI factory used to create the API to interact with `stores`."""
    app = FastAPI(lifespan=lifespan)

    # Give the app access to the project
    if not project:
        project = open("project.skore")

    app.state.project = project

    # Enable CORS support on all routes, for all origins and methods.
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Include routers from bottom to top.
    # Include routers always after all routes have been defined/imported.
    router = APIRouter(prefix="/api")
    router.include_router(project_router)

    # Include all sub routers.
    app.include_router(router)

    # Mount skore-ui from the static directory.
    # Should be after the API routes to avoid shadowing previous routes.
    static_path = get_static_path()
    if static_path.exists():
        # The mimetypes module may fail to set the
        # correct MIME type for javascript files.
        # More info on this here: https://github.com/encode/starlette/issues/829
        # So force it...
        if "win" in sys.platform:
            import mimetypes

            mimetypes.add_type("application/javascript", ".js")

        app.mount(
            "/",
            StaticFiles(
                directory=static_path,
                html=True,
                follow_symlink=True,
            ),
            name="static",
        )
    else:

        async def read_index(request):
            module_path = Path(__file__).resolve().parent
            error_template_path = module_path / "templates" / "404.html"
            return FileResponse(error_template_path, status_code=404)

        app.router.add_route("/", read_index, methods=["GET"])

    return app



"""'Time-based column' warning.

This warning is shown when ``X`` contains a time-based column.
Currently, only pandas and polars DataFrames are supported.
"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Union

from skore.sklearn.train_test_split.warning.train_test_split_warning import (
    TrainTestSplitWarning,
)

if TYPE_CHECKING:
    ArrayLike = Any


class TimeBasedColumnWarning(TrainTestSplitWarning):
    """Check whether the design matrix ``X`` contains a time-based column.

    Currently, only pandas and polars DataFrames are supported.
    """

    @staticmethod
    def _MSG(df_name: str, offending_column_names: list[str]) -> str:
        s = "" if len(offending_column_names) == 1 else "s"

        df_name_info = "" if df_name is None else f', dataframe "{df_name}"'

        column_info = f"(column{s} {', '.join(offending_column_names)}{df_name_info})"

        return (
            f"We detected some time-based columns {column_info} in your data. "
            "We recommend using scikit-learn's TimeSeriesSplit instead of "
            "train_test_split. Otherwise you might train on future data to "
            "predict the past, or get inflated model performance evaluation "
            "because natural drift will not be taken into account."
        )

    @staticmethod
    def check(X: ArrayLike, **kwargs) -> Union[str, None]:
        """Check whether the design matrix ``X`` contains a time-based column.

        Currently, only pandas and polars DataFrames are supported.

        Parameters
        ----------
        X : array-like or None
            A data matrix that can contain datetime data, e.g. a DataFrame.

        Returns
        -------
        warning
            None if the check passed, otherwise the warning message.
        """
        if not hasattr(X, "columns"):
            return None

        dtypes = [(col, X[col].dtype) for col in X.columns]
        # NOTE: Searching by regex allows us to avoid depending explicitly
        # on Pandas or Polars
        datetime_columns = [
            f'"{col}"'
            for col, dtype in dtypes
            if re.search("date", str(type(dtype)), flags=re.IGNORECASE)
        ]
        if datetime_columns:
            df_name = X.name if hasattr(X, "name") else None
            return TimeBasedColumnWarning._MSG(df_name, datetime_columns)



"""SklearnBaseEstimatorItem.

This module defines the SklearnBaseEstimatorItem class,
which represents a scikit-learn BaseEstimator item.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from .item import Item, ItemTypeError

if TYPE_CHECKING:
    import sklearn.base


class SklearnBaseEstimatorItem(Item):
    """
    A class to represent a scikit-learn BaseEstimator item.

    This class encapsulates a scikit-learn BaseEstimator along with its
    creation and update timestamps.
    """

    def __init__(
        self,
        estimator_html_repr: str,
        estimator_skops: bytes,
        estimator_skops_untrusted_types: list[str],
        created_at: Union[str, None] = None,
        updated_at: Union[str, None] = None,
        note: Union[str, None] = None,
    ):
        """
        Initialize a SklearnBaseEstimatorItem.

        Parameters
        ----------
        estimator_html_repr : str
            The HTML representation of the scikit-learn estimator.
        estimator_skops : bytes
            The skops representation of the scikit-learn estimator.
        estimator_skops_untrusted_types : list[str]
            The list of untrusted types in the skops representation.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        note : Union[str, None]
            An optional note.
        """
        super().__init__(created_at, updated_at, note)

        self.estimator_html_repr = estimator_html_repr
        self.estimator_skops = estimator_skops
        self.estimator_skops_untrusted_types = estimator_skops_untrusted_types

    @property
    def estimator(self) -> sklearn.base.BaseEstimator:
        """
        Convert the stored skops object to a scikit-learn BaseEstimator.

        Returns
        -------
        sklearn.base.BaseEstimator
            The scikit-learn BaseEstimator representation of the stored skops object.
        """
        import skops.io

        return skops.io.loads(
            self.estimator_skops, trusted=self.estimator_skops_untrusted_types
        )

    @classmethod
    def factory(
        cls,
        estimator: sklearn.base.BaseEstimator,
        /,
        **kwargs,
    ) -> SklearnBaseEstimatorItem:
        """
        Create a SklearnBaseEstimatorItem instance from a scikit-learn BaseEstimator.

        Parameters
        ----------
        estimator : sklearn.base.BaseEstimator
            The scikit-learn BaseEstimator to store.

        Returns
        -------
        SklearnBaseEstimatorItem
            A new SklearnBaseEstimatorItem instance.
        """
        import sklearn.base
        import sklearn.utils

        if not isinstance(estimator, sklearn.base.BaseEstimator):
            raise ItemTypeError(f"Type '{estimator.__class__}' is not supported.")

        # This line is only needed if we know `estimator` has the right type, so we do
        # it after the type check
        import skops.io

        estimator_html_repr = sklearn.utils.estimator_html_repr(estimator)
        estimator_skops = skops.io.dumps(estimator)
        estimator_skops_untrusted_types = skops.io.get_untrusted_types(
            data=estimator_skops
        )

        return cls(
            estimator_html_repr=estimator_html_repr,
            estimator_skops=estimator_skops,
            estimator_skops_untrusted_types=estimator_skops_untrusted_types,
            **kwargs,
        )

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        return super().as_serializable_dict() | {
            "value": self.estimator_html_repr,
            "media_type": "application/vnd.sklearn.estimator+html",
        }



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import auc, roc_curve
from sklearn.preprocessing import LabelBinarizer

from skore.sklearn._plot.utils import (
    HelpDisplayMixin,
    _ClassifierCurveDisplayMixin,
    _despine_matplotlib_axis,
    _validate_style_kwargs,
    sample_mpl_colormap,
)


class RocCurveDisplay(HelpDisplayMixin, _ClassifierCurveDisplayMixin):
    """ROC Curve visualization.

    An instance of this class is should created by `EstimatorReport.metrics.plot.roc()`.
    You should not create an instance of this class directly.

    Parameters
    ----------
    fpr : dict of list of ndarray
        False positive rate. The structure is:

        - for binary classification:
            - the key is the positive label.
            - the value is a list of `ndarray`, each `ndarray` being the false
              positive rate.
        - for multiclass classification:
            - the key is the class of interest in an OvR fashion.
            - the value is a list of `ndarray`, each `ndarray` being the false
              positive rate.

    tpr : dict of list of ndarray
        True positive rate. The structure is:

        - for binary classification:
            - the key is the positive label
            - the value is a list of `ndarray`, each `ndarray` being the true
              positive rate.
        - for multiclass classification:
            - the key is the class of interest in an OvR fashion.
            - the value is a list of `ndarray`, each `ndarray` being the true
              positive rate.

    roc_auc : dict of list of float
        Area under the ROC curve. The structure is:

        - for binary classification:
            - the key is the positive label
            - the value is a list of `float`, each `float` being the area under
              the ROC curve.
        - for multiclass classification:
            - the key is the class of interest in an OvR fashion.
            - the value is a list of `float`, each `float` being the area under
              the ROC curve.

    estimator_name : str
        Name of the estimator.

    pos_label : int, float, bool or str, default=None
        The class considered as positive. Only meaningful for binary classification.

    data_source : {"train", "test", "X_y"}, default=None
        The data source used to compute the ROC curve.

    Attributes
    ----------
    ax_ : matplotlib axes
        The axes on which the ROC curve is plotted.

    figure_ : matplotlib figure
        The figure on which the ROC curve is plotted.

    lines_ : list of matplotlib lines
        The lines of the ROC curve.

    chance_level_ : matplotlib line
        The chance level line.

    Examples
    --------
    >>> from sklearn.datasets import load_breast_cancer
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import train_test_split
    >>> from skore import EstimatorReport
    >>> X_train, X_test, y_train, y_test = train_test_split(
    ...     *load_breast_cancer(return_X_y=True), random_state=0
    ... )
    >>> classifier = LogisticRegression(max_iter=10_000)
    >>> reporter = EstimatorReport(
    ...     classifier,
    ...     X_train=X_train,
    ...     y_train=y_train,
    ...     X_test=X_test,
    ...     y_test=y_test,
    ... )
    >>> display = reporter.metrics.plot.roc()
    >>> display.plot(roc_curve_kwargs={"color": "tab:red"})
    """

    def __init__(
        self,
        *,
        fpr,
        tpr,
        roc_auc,
        estimator_name,
        pos_label=None,
        data_source=None,
    ):
        self.estimator_name = estimator_name
        self.fpr = fpr
        self.tpr = tpr
        self.roc_auc = roc_auc
        self.pos_label = pos_label
        self.data_source = data_source

    def plot(
        self,
        ax=None,
        *,
        estimator_name=None,
        roc_curve_kwargs=None,
        plot_chance_level=True,
        chance_level_kwargs=None,
        despine=True,
    ):
        """Plot visualization.

        Extra keyword arguments will be passed to matplotlib's ``plot``.

        Parameters
        ----------
        ax : matplotlib axes, default=None
            Axes object to plot on. If `None`, a new figure and axes is
            created.

        estimator_name : str, default=None
            Name of the estimator used to plot the ROC curve. If `None`, we use
            the inferred name from the estimator.

        roc_curve_kwargs : dict or list of dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the ROC curve(s).

        plot_chance_level : bool, default=True
            Whether to plot the chance level.

        chance_level_kwargs : dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the chance level line.

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Returns
        -------
        display : :class:`~sklearn.metrics.RocCurveDisplay`
            Object that stores computed values.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from sklearn.model_selection import train_test_split
        >>> from skore import EstimatorReport
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     *load_breast_cancer(return_X_y=True), random_state=0
        ... )
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> reporter = EstimatorReport(
        ...     classifier,
        ...     X_train=X_train,
        ...     y_train=y_train,
        ...     X_test=X_test,
        ...     y_test=y_test,
        ... )
        >>> display = reporter.metrics.plot.roc()
        >>> display.plot(roc_curve_kwargs={"color": "tab:red"})
        """
        self.ax_, self.figure_, estimator_name = self._validate_plot_params(
            ax=ax, estimator_name=estimator_name
        )

        self.lines_ = []
        if len(self.fpr) == 1:  # binary-classification
            if len(self.fpr[self.pos_label]) == 1:  # single-split
                if roc_curve_kwargs is None:
                    roc_curve_kwargs = {}
                elif isinstance(roc_curve_kwargs, list):
                    if len(roc_curve_kwargs) > 1:
                        raise ValueError(
                            "You intend to plot a single ROC curve and provide "
                            "multiple ROC curve keyword arguments. Provide a single "
                            "dictionary or a list with a single dictionary."
                        )
                    roc_curve_kwargs = roc_curve_kwargs[0]

                fpr = self.fpr[self.pos_label][0]
                tpr = self.tpr[self.pos_label][0]
                roc_auc = self.roc_auc[self.pos_label][0]

                default_line_kwargs = {}
                if self.data_source in ("train", "test"):
                    default_line_kwargs["label"] = (
                        f"{self.data_source.title()} set (AUC = {roc_auc:0.2f})"
                    )
                else:  # data_source in (None, "X_y")
                    default_line_kwargs["label"] = f"AUC = {roc_auc:0.2f}"

                line_kwargs = _validate_style_kwargs(
                    default_line_kwargs, roc_curve_kwargs
                )

                (line_,) = self.ax_.plot(fpr, tpr, **line_kwargs)
                self.lines_.append(line_)
            else:  # cross-validation
                if roc_curve_kwargs is None:
                    roc_curve_kwargs = [{}] * len(self.fpr[self.pos_label])
                elif isinstance(roc_curve_kwargs, dict):
                    roc_curve_kwargs = [roc_curve_kwargs] * len(
                        self.fpr[self.pos_label]
                    )
                elif isinstance(roc_curve_kwargs, list):
                    if len(roc_curve_kwargs) != len(self.fpr[self.pos_label]):
                        raise ValueError(
                            "You intend to plot multiple ROC curves. We expect "
                            "`roc_curve_kwargs` to be a list of dictionaries with the "
                            "same length as the number of ROC curves. Got "
                            f"{len(roc_curve_kwargs)} instead of "
                            f"{len(self.fpr)}."
                        )
                else:
                    raise ValueError(
                        "You intend to plot multiple ROC curves. We expect "
                        "`roc_curve_kwargs` to be a list of dictionaries of "
                        f"{len(self.fpr)} elements. Got {roc_curve_kwargs!r} instead."
                    )

                for split_idx in range(len(self.fpr[self.pos_label])):
                    fpr = self.fpr[self.pos_label][split_idx]
                    tpr = self.tpr[self.pos_label][split_idx]
                    roc_auc = self.roc_auc[self.pos_label][split_idx]

                    default_line_kwargs = {
                        "label": (
                            f"{self.data_source.title()} set - fold #{split_idx + 1} "
                            f"(AUC = {roc_auc:0.2f})"
                        )
                    }
                    line_kwargs = _validate_style_kwargs(
                        default_line_kwargs, roc_curve_kwargs[split_idx]
                    )

                    (line_,) = self.ax_.plot(fpr, tpr, **line_kwargs)
                    self.lines_.append(line_)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )
        else:  # multiclass-classification
            info_pos_label = None  # irrelevant for multiclass
            class_colors = sample_mpl_colormap(
                plt.cm.tab10, 10 if len(self.fpr) < 10 else len(self.fpr)
            )
            if roc_curve_kwargs is None:
                roc_curve_kwargs = [{}] * len(self.fpr)
            elif isinstance(roc_curve_kwargs, list):
                if len(roc_curve_kwargs) != len(self.fpr):
                    raise ValueError(
                        "You intend to plot multiple ROC curves. We expect "
                        "`roc_curve_kwargs` to be a list of dictionaries with the "
                        "same length as the number of ROC curves. Got "
                        f"{len(roc_curve_kwargs)} instead of "
                        f"{len(self.fpr)}."
                    )
            else:
                raise ValueError(
                    "You intend to plot multiple ROC curves. We expect "
                    "`roc_curve_kwargs` to be a list of dictionaries of "
                    f"{len(self.fpr)} elements. Got {roc_curve_kwargs!r} instead."
                )

            for class_idx, class_ in enumerate(self.fpr):
                fpr_class = self.fpr[class_]
                tpr_class = self.tpr[class_]
                roc_auc_class = self.roc_auc[class_]
                roc_curve_kwargs_class = roc_curve_kwargs[class_idx]

                if len(fpr_class) == 1:  # single-split
                    fpr = fpr_class[0]
                    tpr = tpr_class[0]
                    roc_auc = roc_auc_class[0]

                    default_line_kwargs = {"color": class_colors[class_idx]}
                    if self.data_source in ("train", "test"):
                        default_line_kwargs["label"] = (
                            f"{str(class_).title()} - {self.data_source} "
                            f"set (AUC = {roc_auc:0.2f})"
                        )
                    else:  # data_source in (None, "X_y")
                        default_line_kwargs["label"] = (
                            f"{str(class_).title()} - AUC = {roc_auc:0.2f}"
                        )

                    line_kwargs = _validate_style_kwargs(
                        default_line_kwargs, roc_curve_kwargs_class
                    )

                    (line_,) = self.ax_.plot(fpr, tpr, **line_kwargs)
                    self.lines_.append(line_)
                else:  # cross-validation
                    for split_idx in range(len(fpr_class)):
                        fpr = fpr_class[split_idx]
                        tpr = tpr_class[split_idx]
                        roc_auc_mean = np.mean(roc_auc_class)
                        roc_auc_std = np.std(roc_auc_class)

                        default_line_kwargs = {
                            "color": class_colors[class_idx],
                            "alpha": 0.3,
                        }
                        if split_idx == 0:
                            default_line_kwargs["label"] = (
                                f"{str(class_).title()} - {self.data_source} set"
                                f" (AUC = {roc_auc_mean:0.2f} +/- "
                                f"{roc_auc_std:0.2f})"
                            )
                        else:
                            default_line_kwargs["label"] = None

                        line_kwargs = _validate_style_kwargs(default_line_kwargs, {})

                        (line_,) = self.ax_.plot(fpr, tpr, **line_kwargs)
                        self.lines_.append(line_)

        default_chance_level_line_kw = {
            "label": "Chance level (AUC = 0.5)",
            "color": "k",
            "linestyle": "--",
        }

        if chance_level_kwargs is None:
            chance_level_kwargs = {}

        chance_level_kwargs = _validate_style_kwargs(
            default_chance_level_line_kw, chance_level_kwargs
        )

        xlabel = "False Positive Rate"
        ylabel = "True Positive Rate"
        if info_pos_label:
            xlabel += info_pos_label
            ylabel += info_pos_label

        self.ax_.set(
            xlabel=xlabel,
            xlim=(-0.01, 1.01),
            ylabel=ylabel,
            ylim=(-0.01, 1.01),
            aspect="equal",
        )

        if plot_chance_level:
            (self.chance_level_,) = self.ax_.plot((0, 1), (0, 1), **chance_level_kwargs)
        else:
            self.chance_level_ = None

        if despine:
            _despine_matplotlib_axis(self.ax_)

        self.ax_.legend(loc="lower right", title=estimator_name)

    @classmethod
    def _from_predictions(
        cls,
        y_true,
        y_pred,
        *,
        estimator,
        estimator_name,
        ml_task,
        data_source=None,
        pos_label=None,
        drop_intermediate=True,
        ax=None,
        roc_curve_kwargs=None,
        plot_chance_level=True,
        chance_level_kwargs=None,
        despine=True,
    ):
        """Private method to create a RocCurveDisplay from predictions.

        Parameters
        ----------
        y_true : list of array-like of shape (n_samples,)
            True binary labels in binary classification.

        y_pred : list of array-like of shape (n_samples,)
            Target scores, can either be probability estimates of the positive class,
            confidence values, or non-thresholded measure of decisions (as returned by
            “decision_function” on some classifiers).

        estimator : estimator instance
            The estimator from which `y_pred` is obtained.

        estimator_name : str
            Name of the estimator used to plot the ROC curve.

        ml_task : {"binary-classification", "multiclass-classification"}
            The machine learning task.

        data_source : {"train", "test", "X_y"}, default=None
            The data source used to compute the ROC curve.

        pos_label : int, float, bool or str, default=None
            The class considered as the positive class when computing the
            precision and recall metrics.

        drop_intermediate : bool, default=True
            Whether to drop intermediate points with identical value.

        ax : matplotlib axes, default=None
            Axes object to plot on. If `None`, a new figure and axes is
            created.

        roc_curve_kwargs : dict or list of dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the ROC curve(s).

        plot_chance_level : bool, default=True
            Whether to plot the chance level.

        chance_level_kwargs : dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the chance level line.

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Returns
        -------
        display : RocCurveDisplay
            Object that stores computed values.
        """
        pos_label_validated = cls._validate_from_predictions_params(
            y_true, y_pred, ml_task=ml_task, pos_label=pos_label
        )

        if ml_task == "binary-classification":
            fpr, tpr, roc_auc = defaultdict(list), defaultdict(list), defaultdict(list)
            for y_true_i, y_pred_i in zip(y_true, y_pred):
                fpr_i, tpr_i, _ = roc_curve(
                    y_true_i,
                    y_pred_i,
                    pos_label=pos_label,
                    drop_intermediate=drop_intermediate,
                )
                roc_auc_i = auc(fpr_i, tpr_i)
                fpr[pos_label_validated].append(fpr_i)
                tpr[pos_label_validated].append(tpr_i)
                roc_auc[pos_label_validated].append(roc_auc_i)
        else:  # multiclass-classification
            # OvR fashion to collect fpr, tpr, and roc_auc
            fpr, tpr, roc_auc = defaultdict(list), defaultdict(list), defaultdict(list)
            for y_true_i, y_pred_i in zip(y_true, y_pred):
                label_binarizer = LabelBinarizer().fit(estimator.classes_)
                y_true_onehot_i = label_binarizer.transform(y_true_i)
                for class_idx, class_ in enumerate(estimator.classes_):
                    fpr_class_i, tpr_class_i, _ = roc_curve(
                        y_true_onehot_i[:, class_idx],
                        y_pred_i[:, class_idx],
                        pos_label=None,
                        drop_intermediate=drop_intermediate,
                    )
                    roc_auc_class_i = auc(fpr_class_i, tpr_class_i)

                    fpr[class_].append(fpr_class_i)
                    tpr[class_].append(tpr_class_i)
                    roc_auc[class_].append(roc_auc_class_i)

        viz = cls(
            fpr=fpr,
            tpr=tpr,
            roc_auc=roc_auc,
            estimator_name=estimator_name,
            pos_label=pos_label_validated,
            data_source=data_source,
        )

        viz.plot(
            ax=ax,
            estimator_name=estimator_name,
            roc_curve_kwargs=roc_curve_kwargs,
            plot_chance_level=plot_chance_level,
            chance_level_kwargs=chance_level_kwargs,
            despine=despine,
        )

        return viz



"""Define PrimitiveItem.

PrimitiveItems represents a primitive item with creation and update timestamps.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .item import Item, ItemTypeError

if TYPE_CHECKING:
    from typing import Union

    Primitive = Union[
        bool,
        float,
        int,
        str,
        list["Primitive"],
        tuple["Primitive"],
        dict[Union[str, int, float], "Primitive"],
    ]


def is_primitive(obj: object) -> bool:
    """Check if the object is a primitive."""
    if isinstance(obj, (bool, float, int, str)):
        return True
    if isinstance(obj, (list, tuple)):
        return all(is_primitive(item) for item in obj)
    if isinstance(obj, dict):
        return all(
            isinstance(k, (bool, float, int, str)) and is_primitive(v)
            for k, v in obj.items()
        )
    return False


class PrimitiveItem(Item):
    """
    A class to represent a primitive item.

    This class encapsulates a primitive value
    along with its creation and update timestamps.
    """

    def __init__(
        self,
        primitive: Primitive,
        created_at: Union[str, None] = None,
        updated_at: Union[str, None] = None,
        note: Union[str, None] = None,
    ):
        """
        Initialize a PrimitiveItem.

        Parameters
        ----------
        primitive : Primitive
            The primitive value to store.
        created_at : str, optional
            The creation timestamp as ISO format.
        updated_at : str, optional
            The last update timestamp as ISO format.
        note : Union[str, None]
            An optional note.
        """
        super().__init__(created_at, updated_at, note)

        self.primitive = primitive

    @classmethod
    def factory(cls, primitive: Primitive, /, **kwargs) -> PrimitiveItem:
        """
        Create a new PrimitiveItem with the current timestamp.

        Parameters
        ----------
        primitive : Primitive
            The primitive value to store.

        Returns
        -------
        PrimitiveItem
            A new PrimitiveItem instance.
        """
        if isinstance(primitive, str) or not is_primitive(primitive):
            raise ItemTypeError(f"Type '{primitive.__class__}' is not supported.")

        return cls(primitive=primitive, **kwargs)

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        return super().as_serializable_dict() | {
            "media_type": "text/markdown",
            "value": self.primitive,
        }



"""Load project helper."""

from pathlib import Path
from typing import Union

from skore.persistence.repository import ItemRepository, ViewRepository
from skore.persistence.storage.disk_cache_storage import (
    DirectoryDoesNotExist,
    DiskCacheStorage,
)
from skore.project.project import Project


class ProjectLoadError(Exception):
    """Failed to load project."""


def _load(project_name: Union[str, Path]) -> Project:
    """Load an existing Project given a project name or path.

    Transforms a project name to a directory path as follows:
    - Resolves relative path to current working directory,
    - Checks that the file ends with the ".skore" extension,
    - If not provided, it will be automatically appended,
    - If project name is an absolute path, keeps that path.
    """
    path = Path(project_name).resolve()

    if path.suffix != ".skore":
        path = path.parent / (path.name + ".skore")

    if not Path(path).exists():
        raise FileNotFoundError(f"Project '{path}' does not exist: did you create it?")

    try:
        # FIXME: Should those hardcoded strings be factorized somewhere ?
        item_storage = DiskCacheStorage(directory=Path(path) / "items")
        item_repository = ItemRepository(storage=item_storage)
        view_storage = DiskCacheStorage(directory=Path(path) / "views")
        view_repository = ViewRepository(storage=view_storage)
        project = Project(
            item_repository=item_repository,
            view_repository=view_repository,
        )
    except DirectoryDoesNotExist as e:
        missing_directory = e.args[0].split()[1]
        raise ProjectLoadError(
            f"Project '{path}' is corrupted: "
            f"directory '{missing_directory}' should exist. "
            "Consider re-creating the project."
        ) from e

    return project



"""CrossValidationReporterItem.

This module defines the CrossValidationReporterItem class, which is used to persist
reporters of cross-validation.
"""

from __future__ import annotations

import contextlib
import dataclasses
import importlib
import io
import json
import re
import statistics
from typing import TYPE_CHECKING, Literal, Optional, TypedDict

import joblib
import numpy
import plotly.graph_objects
import plotly.io

from skore.sklearn.cross_validation import CrossValidationReporter

from .item import Item, ItemTypeError

if TYPE_CHECKING:
    import sklearn.base

    class EstimatorParamInfo(TypedDict):
        """Information about an estimator parameter."""

        value: str
        default: bool

    class EstimatorInfo(TypedDict):
        """Information about an estimator."""

        name: str
        module: str
        params: dict[str, EstimatorParamInfo]


HUMANIZED_PLOT_NAMES = {
    "scores": "Scores",
    "timing": "Timings",
}


def _metric_title(metric):
    m = metric.replace("_", " ")
    title = f"Mean {m}"
    if title.endswith(" time"):
        title = title + " (seconds)"
    return title


def _metric_favorability(
    metric: str,
) -> Literal["greater_is_better", "lower_is_better", "unknown"]:
    greater_is_better_metrics = (
        "accuracy",
        "balanced_accuracy",
        "top_k_accuracy",
        "average_precision",
        "f1",
        "precision",
        "recall",
        "jaccard",
        "roc_auc",
        "r2",
    )
    any_match_greater_is_better = any(
        re.search(re.escape(pattern), metric) for pattern in greater_is_better_metrics
    )
    if (
        any_match_greater_is_better
        # other scikit-learn conventions
        or metric.endswith("_score")  # score: higher is better
        or metric.startswith("neg_")  # negative loss: negative of lower is better
    ):
        return "greater_is_better"

    lower_is_better_metrics = ("fit_time", "score_time")
    if (
        metric.endswith("_error")
        or metric.endswith("_loss")
        or metric.endswith("_deviance")
        or metric in lower_is_better_metrics
    ):
        return "lower_is_better"

    return "unknown"


def _params_to_str(estimator_info) -> str:
    params_list = []
    for k, v in estimator_info["params"].items():
        value = v["value"]
        if v["default"] is True:
            params_list.append(f"- {k}: {value} (default)")
        else:
            params_list.append(f"- {k}: *{value}*")

    return "\n".join(params_list)


def _estimator_info(estimator: sklearn.base.BaseEstimator) -> EstimatorInfo:
    estimator_params = (
        estimator.get_params() if hasattr(estimator, "get_params") else {}
    )

    name = estimator.__class__.__name__
    module = estimator.__module__

    # Figure out the default parameters of the estimator,
    # so that we can highlight the non-default ones in the UI

    # This is done by instantiating the class with no arguments and
    # computing the diff between the default and ours
    try:
        estimator_module = importlib.import_module(module)
        EstimatorClass = getattr(estimator_module, name)
        default_estimator_params = EstimatorClass().get_params()
    except Exception:
        default_estimator_params = {}

    final_estimator_params: dict[str, EstimatorParamInfo] = {}
    for k, v in estimator_params.items():
        param_is_default: bool = (
            k in default_estimator_params and default_estimator_params[k] == v
        )
        final_estimator_params[str(k)] = {
            "value": repr(v),
            "default": param_is_default,
        }

    return {
        "name": name,
        "module": module,
        "params": final_estimator_params,
    }


class CrossValidationReporterItem(Item):
    """Class to persist the reporter of cross-validation."""

    def __init__(
        self,
        reporter_bytes: bytes,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        note: Optional[str] = None,
    ):
        """
        Initialize a CrossValidationReporterItem.

        Parameters
        ----------
        reporter_bytes : bytes
            The raw bytes of the reporter pickled representation.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        note : str, optional
            A note.
        """
        super().__init__(created_at, updated_at, note)

        self.reporter_bytes = reporter_bytes

    @classmethod
    def factory(
        cls,
        reporter: CrossValidationReporter,
        /,
        **kwargs,
    ) -> CrossValidationReporterItem:
        """
        Create a CrossValidationReporterItem instance from a CrossValidationReporter.

        Parameters
        ----------
        reporter : CrossValidationReporter

        Returns
        -------
        CrossValidationReporterItem
            A new CrossValidationReporterItem instance.
        """
        if not isinstance(reporter, CrossValidationReporter):
            raise ItemTypeError(f"Type '{reporter.__class__}' is not supported.")

        with io.BytesIO() as stream:
            joblib.dump(reporter, stream)

            return cls(stream.getvalue(), **kwargs)

    @property
    def reporter(self) -> CrossValidationReporter:
        """The CrossValidationReporter from the persistence."""
        with io.BytesIO(self.reporter_bytes) as stream:
            return joblib.load(stream)

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        # Get tabular results (the cv results in a dataframe-like structure)
        cv_results = {
            key: value.tolist()
            for key, value in self.reporter._cv_results.items()
            if key not in ("estimator", "indices") and isinstance(value, numpy.ndarray)
        }

        metrics_names = list(cv_results)
        tabular_results = [
            {
                "name": "Cross validation results",
                "columns": metrics_names,
                "data": list(zip(*cv_results.values())),
                "favorability": [_metric_favorability(m) for m in metrics_names],
            }
        ]

        # Get scalar results (summary statistics of the cv results)
        scalar_results = [
            {
                "name": _metric_title(k),
                "value": statistics.mean(v),
                "stddev": statistics.stdev(v),
                "favorability": _metric_favorability(k),
            }
            for k, v in cv_results.items()
        ]

        # If the estimator is from sklearn, make the class name a hyperlink
        # to the relevant docs
        estimator_info = _estimator_info(self.reporter.estimator)
        name = estimator_info["name"]
        module = re.sub(r"\.\_.+", "", estimator_info["module"])
        if module.startswith("sklearn"):
            doc_url = f"https://scikit-learn.org/stable/modules/generated/{module}.{name}.html"
            doc_link = f'<a href="{doc_url}" target="_blank"><code>{name}</code></a>'
        else:
            doc_link = f"`{name}`"

        params_as_str = _params_to_str(estimator_info)
        estimator_params_as_str = f"{doc_link}\n{params_as_str}"

        # Serialize cross-validation settings
        cv = self.reporter.cv
        cv_info: dict[str, str] = {}
        if isinstance(cv, int):
            cv_info["n_splits"] = repr(cv)
        elif cv is None or hasattr(cv, "__iter__"):
            # cv is None or an iterable of splits
            cv_info["n_splits"] = "unknown"
        else:
            # cv is a sklearn CV splitter object
            for attr_name in ["n_splits", "shuffle", "random_state"]:
                with contextlib.suppress(AttributeError):
                    attr = getattr(cv, attr_name)
                    cv_info[attr_name] = repr(attr)

        cv_params_as_str = ", ".join(f"{k}: *{v}*" for k, v in cv_info.items())

        # Construct the final representation
        value = {
            "scalar_results": scalar_results,
            "tabular_results": tabular_results,
            "plots": [
                {
                    "name": HUMANIZED_PLOT_NAMES[plot_name],
                    "value": json.loads(plotly.io.to_json(plot, engine="json")),
                }
                for plot_name, plot in dataclasses.asdict(self.reporter.plots).items()
            ],
            "sections": [
                {
                    "title": "Model",
                    "icon": "icon-square-cursor",
                    "items": [
                        {
                            "name": "Estimator parameters",
                            "description": "Core model configuration used for training",
                            "value": estimator_params_as_str,
                        },
                        {
                            "name": "Cross-validation parameters",
                            "description": "Controls how data is split and validated",
                            "value": cv_params_as_str,
                        },
                    ],
                }
            ],
        }

        return super().as_serializable_dict() | {
            "media_type": "application/vnd.skore.cross_validation+json",
            "value": value,
        }



"""Implement the "quickstart" command."""

from pathlib import Path
from typing import Union

from skore.cli import logger
from skore.cli.launch_dashboard import __launch
from skore.project.create import _create
from skore.utils._logger import logger_context


def __quickstart(
    project_name: Union[str, Path],
    overwrite: bool,
    port: int,
    open_browser: bool,
    verbose: bool = False,
):
    """Quickstart a Skore project.

    Create it if it does not exist, then launch the web UI.

    Parameters
    ----------
    project_name : Path-like
        Name of the project to be created, or a relative or absolute path.
    overwrite : bool
        If ``True``, overwrite an existing project with the same name.
        If ``False``, simply warn that a project already exists.
    port : int
        Port at which to bind the UI server.
    open_browser : bool
        Whether to automatically open a browser tab showing the UI.
    verbose : bool
        Whether to increase logging verbosity.
    """
    with logger_context(logger, verbose):
        try:
            _create(
                project_name=project_name,
                overwrite=overwrite,
                verbose=verbose,
            )
        except FileExistsError:
            logger.info(
                f"Project file '{project_name}' already exists. Skipping creation step."
            )

        path = Path(project_name)

        __launch(
            project_name=path,
            port=port,
            open_browser=open_browser,
            verbose=verbose,
        )



"""Provide a set of classes responsible for manipulating items and views."""

from .item_repository import ItemRepository
from .view_repository import ViewRepository

__all__ = [
    "ItemRepository",
    "ViewRepository",
]



import inspect
from io import StringIO

import matplotlib.pyplot as plt
import numpy as np
from rich.console import Console
from rich.panel import Panel
from rich.tree import Tree
from sklearn.utils.validation import (
    _check_pos_label_consistency,
    check_consistent_length,
)


class HelpDisplayMixin:
    """Mixin class to add help functionality to a class."""

    def _get_attributes_for_help(self):
        """Get the attributes ending with '_' to display in help."""
        attributes = []
        for name in dir(self):
            if name.endswith("_") and not name.startswith("_"):
                attributes.append(f".{name}")
        return sorted(attributes)

    def _get_methods_for_help(self):
        """Get the public methods to display in help."""
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        filtered_methods = []
        for name, method in methods:
            is_private = name.startswith("_")
            is_class_method = inspect.ismethod(method) and method.__self__ is type(self)
            is_help_method = name == "help"
            if not (is_private or is_class_method or is_help_method):
                filtered_methods.append((f".{name}(...)", method))
        return sorted(filtered_methods)

    def _create_help_tree(self):
        """Create a rich Tree with attributes and methods."""
        tree = Tree("display")

        attributes = self._get_attributes_for_help()
        attr_branch = tree.add("[bold cyan] Attributes[/bold cyan]")
        # Ensure figure_ and ax_ are first
        sorted_attrs = sorted(attributes)
        sorted_attrs.remove(".ax_")
        sorted_attrs.remove(".figure_")
        sorted_attrs = [".figure_", ".ax_"] + [
            attr for attr in sorted_attrs if attr not in [".figure_", ".ax_"]
        ]
        for attr in sorted_attrs:
            attr_branch.add(attr)

        methods = self._get_methods_for_help()
        method_branch = tree.add("[bold cyan]Methods[/bold cyan]")
        for name, method in methods:
            description = (
                method.__doc__.split("\n")[0]
                if method.__doc__
                else "No description available"
            )
            method_branch.add(f"{name} - {description}")

        return tree

    def _create_help_panel(self):
        return Panel(
            self._create_help_tree(),
            title=(
                f"[bold cyan]{self.__class__.__name__} for "
                f"{self.estimator_name}[/bold cyan]"
            ),
            border_style="orange1",
            expand=False,
        )

    def help(self):
        """Display available attributes and methods using rich."""
        from skore import console  # avoid circular import

        console.print(self._create_help_panel())

    def __str__(self):
        """Return a string representation using rich."""
        console = Console(file=StringIO(), force_terminal=False)
        console.print(
            Panel(
                "Get guidance using the display.help() method",
                title=f"[cyan]{self.__class__.__name__}[/cyan]",
                border_style="orange1",
                expand=False,
            )
        )
        return console.file.getvalue()

    def __repr__(self):
        """Return a string representation using rich."""
        console = Console(file=StringIO(), force_terminal=False)
        console.print(f"[cyan]skore.{self.__class__.__name__}(...)[/cyan]")
        return console.file.getvalue()


class _ClassifierCurveDisplayMixin:
    """Mixin class to be used in Displays requiring a binary classifier.

    The aim of this class is to centralize some validations regarding the estimator and
    the target and gather the response of the estimator.
    """

    def _validate_plot_params(self, *, ax, estimator_name):
        if ax is None:
            _, ax = plt.subplots()

        estimator_name = (
            self.estimator_name if estimator_name is None else estimator_name
        )
        return ax, ax.figure, estimator_name

    @classmethod
    def _validate_from_predictions_params(
        cls,
        y_true,
        y_pred,
        *,
        ml_task,
        pos_label=None,
    ):
        for y_true_i, y_pred_i in zip(y_true, y_pred):
            check_consistent_length(y_true_i, y_pred_i)

        if ml_task == "binary-classification":
            pos_label = _check_pos_label_consistency(pos_label, y_true[0])

        return pos_label


def _despine_matplotlib_axis(ax, *, x_range=(0, 1), y_range=(0, 1), offset=10):
    """Despine the matplotlib axis.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The matplotlib axis to despine.
    x_range : tuple of float, default=(0, 1)
        The range of the x-axis.
    y_range : tuple of float, default=(0, 1)
        The range of the y-axis.
    offset : float, default=10
        The offset to add to the x-axis and y-axis.
    """
    for s in ["top", "right"]:
        ax.spines[s].set_visible(False)
    ax.spines["bottom"].set_bounds(x_range[0], x_range[1])
    ax.spines["left"].set_bounds(y_range[0], y_range[1])
    ax.spines["left"].set_position(("outward", offset))
    ax.spines["bottom"].set_position(("outward", offset))


def _validate_style_kwargs(default_style_kwargs, user_style_kwargs):
    """Create valid style kwargs by avoiding Matplotlib alias errors.

    Matplotlib raises an error when, for example, 'color' and 'c', or 'linestyle' and
    'ls', are specified together. To avoid this, we automatically keep only the one
    specified by the user and raise an error if the user specifies both.

    Parameters
    ----------
    default_style_kwargs : dict
        The Matplotlib style kwargs used by default in the scikit-learn display.
    user_style_kwargs : dict
        The user-defined Matplotlib style kwargs.

    Returns
    -------
    valid_style_kwargs : dict
        The validated style kwargs taking into account both default and user-defined
        Matplotlib style kwargs.
    """
    invalid_to_valid_kw = {
        "ls": "linestyle",
        "c": "color",
        "ec": "edgecolor",
        "fc": "facecolor",
        "lw": "linewidth",
        "mec": "markeredgecolor",
        "mfcalt": "markerfacecoloralt",
        "ms": "markersize",
        "mew": "markeredgewidth",
        "mfc": "markerfacecolor",
        "aa": "antialiased",
        "ds": "drawstyle",
        "font": "fontproperties",
        "family": "fontfamily",
        "name": "fontname",
        "size": "fontsize",
        "stretch": "fontstretch",
        "style": "fontstyle",
        "variant": "fontvariant",
        "weight": "fontweight",
        "ha": "horizontalalignment",
        "va": "verticalalignment",
        "ma": "multialignment",
    }
    for invalid_key, valid_key in invalid_to_valid_kw.items():
        if invalid_key in user_style_kwargs and valid_key in user_style_kwargs:
            raise TypeError(
                f"Got both {invalid_key} and {valid_key}, which are aliases of one "
                "another"
            )
    valid_style_kwargs = default_style_kwargs.copy()

    for key in user_style_kwargs:
        if key in invalid_to_valid_kw:
            valid_style_kwargs[invalid_to_valid_kw[key]] = user_style_kwargs[key]
        else:
            valid_style_kwargs[key] = user_style_kwargs[key]

    return valid_style_kwargs


def sample_mpl_colormap(cmap, n):
    """Sample colors from a Matplotlib colormap.

    Parameters
    ----------
    cmap : matplotlib.colors.Colormap
        The Matplotlib colormap to sample from.
    n : int
        The number of colors to sample.

    Returns
    -------
    colors : list of str
        The sampled colors.
    """
    indices = np.linspace(0, 1, n)
    return [cmap(i) for i in indices]



"""Train-test split.

This module implements `train_test_split`, a wrapper over scikit-learn's own
function of the same name. This wrapper adds some functionality in the form
of warnings shown to the user when common issues are detected, e.g. imbalanced
data, time-series data...
"""



"""PandasDataFrameItem.

This module defines the PandasDataFrameItem class,
which represents a pandas DataFrame item.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from .item import Item, ItemTypeError

if TYPE_CHECKING:
    import pandas


class PandasDataFrameItem(Item):
    """
    A class to represent a pandas DataFrame item.

    This class encapsulates a pandas DataFrame along with its
    creation and update timestamps.
    """

    ORIENT = "split"

    def __init__(
        self,
        index_json: str,
        dataframe_json: str,
        created_at: Union[str, None] = None,
        updated_at: Union[str, None] = None,
        note: Union[str, None] = None,
    ):
        """
        Initialize a PandasDataFrameItem.

        Parameters
        ----------
        index_json : str
            The JSON representation of the dataframe's index.
        dataframe_json : str
            The JSON representation of the dataframe, without its index.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        note : Union[str, None]
            An optional note.
        """
        super().__init__(created_at, updated_at, note)

        self.index_json = index_json
        self.dataframe_json = dataframe_json

    @property
    def dataframe(self) -> pandas.DataFrame:
        """
        The pandas DataFrame from the persistence.

        Its content can differ from the original dataframe because it has been
        serialized using pandas' `to_json` function and not pickled, in order to be
        environment-independent.
        """
        import io

        import pandas

        with (
            io.StringIO(self.index_json) as index_stream,
            io.StringIO(self.dataframe_json) as df_stream,
        ):
            index = pandas.read_json(index_stream, orient=self.ORIENT, dtype=False)
            index = index.set_index(list(index.columns))
            dataframe = pandas.read_json(df_stream, orient=self.ORIENT, dtype=False)
            dataframe.index = index.index

            return dataframe

    @classmethod
    def factory(cls, dataframe: pandas.DataFrame, /, **kwargs) -> PandasDataFrameItem:
        """
        Create a new PandasDataFrameItem instance from a pandas DataFrame.

        Parameters
        ----------
        dataframe : pd.DataFrame
            The pandas DataFrame to store.

        Returns
        -------
        PandasDataFrameItem
            A new PandasDataFrameItem instance.

        Notes
        -----
        The dataframe must be JSON serializable.
        """
        import pandas

        if not isinstance(dataframe, pandas.DataFrame):
            raise ItemTypeError(f"Type '{dataframe.__class__}' is not supported.")

        # Two native methods are available to serialize dataframe with multi-index,
        # while keeping the index names:
        #
        # 1. Using table orientation with JSON serializer:
        #    ```python
        #    json = dataframe.to_json(orient="table")
        #    dataframe = pandas.read_json(json, orient="table", dtype=False)
        #    ```
        #
        #    This method fails when an index/column name is an integer.
        #
        # 2. Using record orientation with indexes as columns:
        #    ```python
        #    dataframe = dataframe.reset_index()
        #    json = dataframe.to_json(orient="records")
        #    dataframe = pandas.read_json(json, orient="records", dtype=False)
        #    ```
        #
        #    This method fails when the index has the same name as one of the columns.
        #
        # None of those methods being compatible, we decide to store indexes separately.

        index = dataframe.index.to_frame(index=False)
        dataframe = dataframe.reset_index(drop=True)

        return cls(
            index_json=index.to_json(orient=PandasDataFrameItem.ORIENT),
            dataframe_json=dataframe.to_json(orient=PandasDataFrameItem.ORIENT),
            **kwargs,
        )

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        return super().as_serializable_dict() | {
            "media_type": "application/vnd.dataframe",
            "value": self.dataframe.fillna("NaN").to_dict(orient="tight"),
        }



from skore.sklearn._plot.precision_recall_curve import PrecisionRecallCurveDisplay
from skore.sklearn._plot.prediction_error import PredictionErrorDisplay
from skore.sklearn._plot.roc_curve import RocCurveDisplay

__all__ = [
    "RocCurveDisplay",
    "PrecisionRecallCurveDisplay",
    "PredictionErrorDisplay",
]



"""MatplotlibFigureItem.

This module defines the MatplotlibFigureItem class, used to persist Matplotlib figures.
"""

from __future__ import annotations

from base64 import b64encode
from io import BytesIO
from typing import TYPE_CHECKING, Optional

import joblib

from .item import Item, ItemTypeError
from .media_item import lazy_is_instance

if TYPE_CHECKING:
    from matplotlib.figure import Figure


class MatplotlibFigureItem(Item):
    """A class used to persist a Matplotlib figure."""

    def __init__(
        self,
        figure_bytes: str,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        note: Optional[str] = None,
    ):
        """
        Initialize a MatplotlibFigureItem.

        Parameters
        ----------
        figure_bytes : bytes
            The raw bytes of the Matplotlib figure pickled representation.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        note : str, optional
            A note.
        """
        super().__init__(created_at, updated_at, note)

        self.figure_bytes = figure_bytes

    @classmethod
    def factory(cls, figure: Figure, /, **kwargs) -> MatplotlibFigureItem:
        """
        Create a new MatplotlibFigureItem instance from a Matplotlib figure.

        Parameters
        ----------
        figure : matplotlib.figure.Figure
            The Matplotlib figure to store.

        Returns
        -------
        MatplotlibFigureItem
            A new MatplotlibFigureItem instance.
        """
        if not lazy_is_instance(figure, "matplotlib.figure.Figure"):
            raise ItemTypeError(f"Type '{figure.__class__}' is not supported.")

        with BytesIO() as stream:
            joblib.dump(figure, stream)

            return cls(stream.getvalue(), **kwargs)

    @property
    def figure(self) -> Figure:
        """The figure from the persistence."""
        with BytesIO(self.figure_bytes) as stream:
            return joblib.load(stream)

    def as_serializable_dict(self) -> dict:
        """Convert item to a JSON-serializable dict to used by frontend."""
        with BytesIO() as stream:
            self.figure.savefig(stream, format="svg", bbox_inches="tight")

            figure_bytes = stream.getvalue()
            figure_b64_str = b64encode(figure_bytes).decode()

            return super().as_serializable_dict() | {
                "media_type": "image/svg+xml;base64",
                "value": figure_b64_str,
            }



"""Custom exceptions."""


class InvalidProjectNameError(Exception):
    """The project name does not fit with one or more of the project name rules.

    - The project name must start with an alphanumeric character, and must not contain
    special characters other than '_' (underscore) and '-' (hyphen).
    - The project name must be at most 255 characters long (including ".skore").
    - The project name must not be a reserved OS file name.
    For example, CON, AUX, NUL... on Windows.
    """


class ProjectCreationError(Exception):
    """Project creation failed."""


class ProjectPermissionError(Exception):
    """Permissions in the directory do not allow creating a file."""



"""Train-test split warning.

This module defines an interface for warnings shown in :func:`~skore.train_test_split`.
"""

from typing import Union


class TrainTestSplitWarning(Warning):
    """Interface for a train-test-split warning."""

    @staticmethod
    def check(*args, **kwargs) -> Union[str, None]:
        """Perform the check.

        Returns
        -------
        warning
            None if the check passed, otherwise the warning message.
        """
        ...



"""Implement skore CLI."""

import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())  # Default to no output
logger.setLevel(logging.INFO)
logger.propagate = False



"""AltairChartItem.

This module defines the AltairChartItem class, used to persist Altair Charts.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .item import Item, ItemTypeError
from .media_item import lazy_is_instance

if TYPE_CHECKING:
    from altair.vegalite.v5.schema.core import TopLevelSpec as AltairChart


class AltairChartItem(Item):
    """A class used to persist a Altair chart."""

    def __init__(
        self,
        chart_str: str,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        note: Optional[str] = None,
    ):
        """
        Initialize a AltairChartItem.

        Parameters
        ----------
        chart_str : bytes
            The JSON str of the Altair chart.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        note : str, optional
            A note.
        """
        super().__init__(created_at, updated_at, note)

        self.chart_str = chart_str

    @classmethod
    def factory(cls, chart: AltairChart, /, **kwargs) -> AltairChartItem:
        """
        Create a new AltairChartItem instance from a Altair chart.

        Parameters
        ----------
        chart : altair.vegalite.v5.schema.core.TopLevelSpec
            The Altair chart to store.

        Returns
        -------
        AltairChartItem
            A new AltairChartItem instance.
        """
        if not lazy_is_instance(chart, "altair.vegalite.v5.schema.core.TopLevelSpec"):
            raise ItemTypeError(f"Type '{chart.__class__}' is not supported.")

        return cls(chart.to_json(), **kwargs)

    @property
    def chart(self) -> AltairChart:
        """The chart from the persistence."""
        import altair

        return altair.Chart.from_json(self.chart_str)

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        import base64

        chart_bytes = self.chart_str.encode("utf-8")
        chart_b64_str = base64.b64encode(chart_bytes).decode()

        return super().as_serializable_dict() | {
            "media_type": "application/vnd.vega.v5+json;base64",
            "value": chart_b64_str,
        }



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import average_precision_score, precision_recall_curve
from sklearn.preprocessing import LabelBinarizer

from skore.sklearn._plot.utils import (
    HelpDisplayMixin,
    _ClassifierCurveDisplayMixin,
    _despine_matplotlib_axis,
    _validate_style_kwargs,
    sample_mpl_colormap,
)


class PrecisionRecallCurveDisplay(HelpDisplayMixin, _ClassifierCurveDisplayMixin):
    """Precision Recall visualization.

    An instance of this class is should created by
    `EstimatorReport.metrics.plot.precision_recall()`. You should not create an
    instance of this class directly.


    Parameters
    ----------
    precision : dict of list of ndarray
        Precision values. The structure is:

        - for binary classification:
            - the key is the positive label.
            - the value is a list of `ndarray`, each `ndarray` being the precision.
        - for multiclass classification:
            - the key is the class of interest in an OvR fashion.
            - the value is a list of `ndarray`, each `ndarray` being the precision.

    recall : dict of list of ndarray
        Recall values. The structure is:

        - for binary classification:
            - the key is the positive label.
            - the value is a list of `ndarray`, each `ndarray` being the recall.
        - for multiclass classification:
            - the key is the class of interest in an OvR fashion.
            - the value is a list of `ndarray`, each `ndarray` being the recall.

    average_precision : dict of list of float
        Average precision. The structure is:

        - for binary classification:
            - the key is the positive label.
            - the value is a list of `float`, each `float` being the average
              precision.
        - for multiclass classification:
            - the key is the class of interest in an OvR fashion.
            - the value is a list of `float`, each `float` being the average
              precision.

    estimator_name : str
        Name of the estimator.

    pos_label : int, float, bool or str, default=None
        The class considered as the positive class. If None, the class will not
        be shown in the legend.

    data_source : {"train", "test", "X_y"}, default=None
        The data source used to compute the precision recall curve.

    Attributes
    ----------
    ax_ : matplotlib Axes
        Axes with precision recall curve.

    figure_ : matplotlib Figure
        Figure containing the curve.

    lines_ : list of matplotlib Artist
        Precision recall curve.

    chance_levels_ : matplotlib Artist or None
        The chance level line. It is `None` if the chance level is not plotted.

    Examples
    --------
    >>> from sklearn.datasets import load_breast_cancer
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import train_test_split
    >>> from skore import EstimatorReport
    >>> X_train, X_test, y_train, y_test = train_test_split(
    ...     *load_breast_cancer(return_X_y=True), random_state=0
    ... )
    >>> classifier = LogisticRegression(max_iter=10_000)
    >>> reporter = EstimatorReport(
    ...     classifier,
    ...     X_train=X_train,
    ...     y_train=y_train,
    ...     X_test=X_test,
    ...     y_test=y_test,
    ... )
    >>> display = reporter.metrics.plot.precision_recall()
    >>> display.plot(pr_curve_kwargs={"color": "tab:red"})
    """

    def __init__(
        self,
        precision,
        recall,
        *,
        average_precision,
        estimator_name,
        pos_label=None,
        data_source=None,
    ):
        self.precision = precision
        self.recall = recall
        self.average_precision = average_precision
        self.estimator_name = estimator_name
        self.pos_label = pos_label
        self.data_source = data_source

    def plot(
        self,
        ax=None,
        *,
        estimator_name=None,
        pr_curve_kwargs=None,
        despine=True,
    ):
        """Plot visualization.

        Extra keyword arguments will be passed to matplotlib's `plot`.

        Parameters
        ----------
        ax : Matplotlib Axes, default=None
            Axes object to plot on. If `None`, a new figure and axes is
            created.

        estimator_name : str, default=None
            Name of the estimator used to plot the precision-recall curve. If
            `None`, we use the inferred name from the estimator.

        pr_curve_kwargs : dict or list of dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the precision-recall curve(s).

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Returns
        -------
        display : PrecisionRecallCurveDisplay
            Object that stores computed values.

        Notes
        -----
        The average precision (cf. :func:`~sklearn.metrics.average_precision_score`)
        in scikit-learn is computed without any interpolation. To be consistent
        with this metric, the precision-recall curve is plotted without any
        interpolation as well (step-wise style).

        You can change this style by passing the keyword argument
        `drawstyle="default"`. However, the curve will not be strictly
        consistent with the reported average precision.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from sklearn.model_selection import train_test_split
        >>> from skore import EstimatorReport
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     *load_breast_cancer(return_X_y=True), random_state=0
        ... )
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> reporter = EstimatorReport(
        ...     classifier,
        ...     X_train=X_train,
        ...     y_train=y_train,
        ...     X_test=X_test,
        ...     y_test=y_test,
        ... )
        >>> display = reporter.metrics.plot.precision_recall()
        >>> display.plot(pr_curve_kwargs={"color": "tab:red"})
        """
        self.ax_, self.figure_, estimator_name = self._validate_plot_params(
            ax=ax, estimator_name=estimator_name
        )

        self.lines_ = []
        self.chance_levels_ = []
        if len(self.precision) == 1:  # binary-classification
            if len(self.precision[self.pos_label]) == 1:  # single-split
                if pr_curve_kwargs is None:
                    pr_curve_kwargs = {}
                elif isinstance(pr_curve_kwargs, list):
                    if len(pr_curve_kwargs) > 1:
                        raise ValueError(
                            "You intend to plot a single precision-recall curve and "
                            "provide multiple precision-recall curve keyword "
                            "arguments. Provide a single dictionary or a list with "
                            "a single dictionary."
                        )
                    pr_curve_kwargs = pr_curve_kwargs[0]

                precision = self.precision[self.pos_label][0]
                recall = self.recall[self.pos_label][0]
                average_precision = self.average_precision[self.pos_label][0]

                default_line_kwargs = {"drawstyle": "steps-post"}
                if self.data_source in ("train", "test"):
                    default_line_kwargs["label"] = (
                        f"{self.data_source.title()} set "
                        f"(AP = {average_precision:0.2f})"
                    )
                else:  # data_source in (None, "X_y")
                    default_line_kwargs["label"] = f"AP = {average_precision:0.2f}"

                line_kwargs = _validate_style_kwargs(
                    default_line_kwargs, pr_curve_kwargs
                )

                (line_,) = self.ax_.plot(recall, precision, **line_kwargs)
                self.lines_.append(line_)
            else:  # cross-validation
                if pr_curve_kwargs is None:
                    pr_curve_kwargs = [{}] * len(self.precision[self.pos_label])
                elif isinstance(pr_curve_kwargs, list):
                    if len(pr_curve_kwargs) != len(self.precision[self.pos_label]):
                        raise ValueError(
                            "You intend to plot multiple precision-recall curves. We "
                            "expect `pr_curve_kwargs` to be a list of dictionaries "
                            "with the same length as the number of precision-recall "
                            "curves. Got "
                            f"{len(pr_curve_kwargs)} instead of "
                            f"{len(self.precision)}."
                        )
                else:
                    raise ValueError(
                        "You intend to plot multiple precision-recall curves. We "
                        "expect `pr_curve_kwargs` to be a list of dictionaries of "
                        f"{len(self.precision)} elements. Got {pr_curve_kwargs!r} "
                        "instead."
                    )

                for split_idx in range(len(self.precision[self.pos_label])):
                    precision = self.precision[self.pos_label][split_idx]
                    recall = self.recall[self.pos_label][split_idx]
                    average_precision = self.average_precision[self.pos_label][
                        split_idx
                    ]

                    default_line_kwargs = {
                        "drawstyle": "steps-post",
                        "label": (
                            f"{self.data_source.title()} set - fold #{split_idx + 1} "
                            f"(AP = {average_precision:0.2f})"
                        ),
                    }
                    line_kwargs = _validate_style_kwargs(
                        default_line_kwargs, pr_curve_kwargs[split_idx]
                    )

                    (line_,) = self.ax_.plot(recall, precision, **line_kwargs)
                    self.lines_.append(line_)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )
        else:  # multiclass-classification
            info_pos_label = None  # irrelevant for multiclass
            class_colors = sample_mpl_colormap(
                plt.cm.tab10, 10 if len(self.precision) < 10 else len(self.precision)
            )
            if pr_curve_kwargs is None:
                pr_curve_kwargs = [{}] * len(self.precision)
            elif isinstance(pr_curve_kwargs, list):
                if len(pr_curve_kwargs) != len(self.precision):
                    raise ValueError(
                        "You intend to plot multiple precision-recall curves. We "
                        "expect `pr_curve_kwargs` to be a list of dictionaries with "
                        "the same length as the number of precision-recall curves. "
                        "Got "
                        f"{len(pr_curve_kwargs)} instead of "
                        f"{len(self.precision)}."
                    )
            else:
                raise ValueError(
                    "You intend to plot multiple precision-recall curves. We expect "
                    "`pr_curve_kwargs` to be a list of dictionaries of "
                    f"{len(self.precision)} elements. Got {pr_curve_kwargs!r} instead."
                )

            for class_idx, class_ in enumerate(self.precision):
                precision_class = self.precision[class_]
                recall_class = self.recall[class_]
                average_precision_class = self.average_precision[class_]
                pr_curve_kwargs_class = pr_curve_kwargs[class_idx]

                if len(precision_class) == 1:  # single-split
                    precision = precision_class[0]
                    recall = recall_class[0]
                    average_precision = average_precision_class[0]

                    default_line_kwargs = {
                        "drawstyle": "steps-post",
                        "color": class_colors[class_idx],
                    }
                    if self.data_source in ("train", "test"):
                        default_line_kwargs["label"] = (
                            f"{str(class_).title()} - {self.data_source} set "
                            f"(AP = {average_precision:0.2f})"
                        )
                    else:  # data_source in (None, "X_y")
                        default_line_kwargs["label"] = (
                            f"{str(class_).title()} AP = {average_precision:0.2f}"
                        )

                    line_kwargs = _validate_style_kwargs(
                        default_line_kwargs, pr_curve_kwargs_class
                    )

                    (line_,) = self.ax_.plot(recall, precision, **line_kwargs)
                    self.lines_.append(line_)
                else:  # cross-validation
                    for split_idx in range(len(precision_class)):
                        precision = precision_class[split_idx]
                        recall = recall_class[split_idx]
                        average_precision = average_precision_class[split_idx]

                        default_line_kwargs = {
                            "color": class_colors[class_idx],
                            "alpha": 0.3,
                        }
                        if split_idx == 0:
                            default_line_kwargs["label"] = (
                                f"{str(class_).title()} - {self.data_source} set"
                                f" (AP = {np.mean(average_precision_class):0.2f} +/- "
                                f"{np.std(average_precision_class):0.2f})"
                            )
                        else:
                            default_line_kwargs["label"] = None

                        line_kwargs = _validate_style_kwargs(default_line_kwargs, {})

                        (line_,) = self.ax_.plot(recall, precision, **line_kwargs)
                        self.lines_.append(line_)

        xlabel = "Recall"
        ylabel = "Precision"
        if info_pos_label:
            xlabel += info_pos_label
            ylabel += info_pos_label

        self.ax_.set(
            xlabel=xlabel,
            xlim=(-0.01, 1.01),
            ylabel=ylabel,
            ylim=(-0.01, 1.01),
            aspect="equal",
        )

        if despine:
            _despine_matplotlib_axis(self.ax_)

        self.ax_.legend(loc="lower left", title=estimator_name)

    @classmethod
    def _from_predictions(
        cls,
        y_true,
        y_pred,
        *,
        estimator,
        estimator_name,
        ml_task,
        data_source=None,
        pos_label=None,
        drop_intermediate=False,
        ax=None,
        pr_curve_kwargs=None,
        despine=True,
    ):
        """Plot precision-recall curve given binary class predictions.

        Parameters
        ----------
        y_true : array-like of shape (n_samples,)
            True binary labels.

        y_pred : array-like of shape (n_samples,)
            Target scores, can either be probability estimates of the positive class,
            confidence values, or non-thresholded measure of decisions (as returned by
            “decision_function” on some classifiers).

        estimator : estimator instance
            The estimator from which `y_pred` is obtained.

        estimator_name : str
            Name of the estimator used to plot the precision-recall curve.

        ml_task : {"binary-classification", "multiclass-classification"}
            The machine learning task.

        data_source : {"train", "test", "X_y"}, default=None
            The data source used to compute the ROC curve.

        pos_label : int, float, bool or str, default=None
            The class considered as the positive class when computing the
            precision and recall metrics.

        drop_intermediate : bool, default=False
            Whether to drop some suboptimal thresholds which would not appear
            on a plotted precision-recall curve. This is useful in order to
            create lighter precision-recall curves.

        ax : matplotlib axes, default=None
            Axes object to plot on. If `None`, a new figure and axes is created.

        pr_curve_kwargs : dict or list of dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the precision-recall curve(s).

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        **kwargs : dict
            Keyword arguments to be passed to matplotlib's `plot`.

        Returns
        -------
        display : :class:`~sklearn.metrics.PrecisionRecallDisplay`
        """
        pos_label_validated = cls._validate_from_predictions_params(
            y_true, y_pred, ml_task=ml_task, pos_label=pos_label
        )

        if ml_task == "binary-classification":
            precision, recall = defaultdict(list), defaultdict(list)
            average_precision = defaultdict(list)
            for y_true_i, y_pred_i in zip(y_true, y_pred):
                precision_i, recall_i, _ = precision_recall_curve(
                    y_true_i,
                    y_pred_i,
                    pos_label=pos_label_validated,
                    drop_intermediate=drop_intermediate,
                )
                average_precision_i = average_precision_score(
                    y_true_i, y_pred_i, pos_label=pos_label_validated
                )

                precision[pos_label_validated].append(precision_i)
                recall[pos_label_validated].append(recall_i)
                average_precision[pos_label_validated].append(average_precision_i)
        else:  # multiclass-classification
            precision, recall = defaultdict(list), defaultdict(list)
            average_precision = defaultdict(list)
            for y_true_i, y_pred_i in zip(y_true, y_pred):
                label_binarizer = LabelBinarizer().fit(estimator.classes_)
                y_true_onehot_i = label_binarizer.transform(y_true_i)
                for class_idx, class_ in enumerate(estimator.classes_):
                    precision_class_i, recall_class_i, _ = precision_recall_curve(
                        y_true_onehot_i[:, class_idx],
                        y_pred_i[:, class_idx],
                        pos_label=None,
                        drop_intermediate=drop_intermediate,
                    )
                    average_precision_class_i = average_precision_score(
                        y_true_onehot_i[:, class_idx], y_pred_i[:, class_idx]
                    )

                    precision[class_].append(precision_class_i)
                    recall[class_].append(recall_class_i)
                    average_precision[class_].append(average_precision_class_i)

        viz = cls(
            precision=precision,
            recall=recall,
            average_precision=average_precision,
            estimator_name=estimator_name,
            pos_label=pos_label_validated,
            data_source=data_source,
        )

        viz.plot(
            ax=ax,
            estimator_name=estimator_name,
            pr_curve_kwargs=pr_curve_kwargs,
            despine=despine,
        )

        return viz



from typing import Any, Callable, Literal, Optional, Union

import matplotlib.axes
import numpy as np
import pandas as pd
from sklearn.metrics import PrecisionRecallDisplay, RocCurveDisplay

from skore.sklearn._base import _BaseAccessor
from skore.sklearn._plot import PredictionErrorDisplay

class _PlotMetricsAccessor(_BaseAccessor):
    _metrics_parent: _MetricsAccessor

    def __init__(self, parent: _MetricsAccessor) -> None: ...
    def _get_display(
        self,
        *,
        X: Optional[np.ndarray],
        y: Optional[np.ndarray],
        data_source: Literal["test", "train", "X_y"],
        response_method: Union[str, list[str]],
        display_class: Any,
        display_kwargs: dict[str, Any],
        display_plot_kwargs: dict[str, Any],
    ) -> Union[RocCurveDisplay, PrecisionRecallDisplay, PredictionErrorDisplay]: ...
    def roc(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        pos_label: Optional[Union[str, int]] = None,
        ax: Optional[matplotlib.axes.Axes] = None,
    ) -> RocCurveDisplay: ...
    def precision_recall(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        pos_label: Optional[Union[str, int]] = None,
        ax: Optional[matplotlib.axes.Axes] = None,
    ) -> PrecisionRecallDisplay: ...
    def prediction_error(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        ax: Optional[matplotlib.axes.Axes] = None,
        kind: Literal[
            "actual_vs_predicted", "residual_vs_predicted"
        ] = "residual_vs_predicted",
        subsample: Optional[Union[int, float]] = 1_000,
    ) -> PredictionErrorDisplay: ...

class _MetricsAccessor(_BaseAccessor):
    _SCORE_OR_LOSS_ICONS: dict[str, str]
    plot: _PlotMetricsAccessor

    def _compute_metric_scores(
        self,
        metric_fn: Callable,
        X: Optional[np.ndarray],
        y_true: Optional[np.ndarray],
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        response_method: Union[str, list[str]],
        pos_label: Optional[Union[str, int]] = None,
        metric_name: Optional[str] = None,
        **metric_kwargs: Any,
    ) -> pd.DataFrame: ...
    def report_metrics(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        scoring: Optional[Union[list[str], Callable]] = None,
        pos_label: Optional[Union[str, int]] = None,
        scoring_kwargs: Optional[dict[str, Any]] = None,
    ) -> pd.DataFrame: ...
    def accuracy(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
    ) -> pd.DataFrame: ...
    def precision(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        average: Optional[
            Literal["binary", "micro", "macro", "weighted", "samples"]
        ] = None,
        pos_label: Optional[Union[str, int]] = None,
    ) -> pd.DataFrame: ...
    def recall(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        average: Optional[
            Literal["binary", "micro", "macro", "weighted", "samples"]
        ] = None,
        pos_label: Optional[Union[str, int]] = None,
    ) -> pd.DataFrame: ...
    def brier_score(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        pos_label: Optional[Union[str, int]] = None,
    ) -> pd.DataFrame: ...
    def roc_auc(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        average: Optional[
            Literal["auto", "micro", "macro", "weighted", "samples"]
        ] = None,
        multi_class: Literal["raise", "ovr", "ovo"] = "ovr",
    ) -> pd.DataFrame: ...
    def log_loss(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
    ) -> pd.DataFrame: ...
    def r2(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        multioutput: Union[
            Literal["raw_values", "uniform_average"], np.ndarray
        ] = "raw_values",
    ) -> pd.DataFrame: ...
    def rmse(
        self,
        *,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        multioutput: Union[
            Literal["raw_values", "uniform_average"], np.ndarray
        ] = "raw_values",
    ) -> pd.DataFrame: ...
    def custom_metric(
        self,
        metric_function: Callable,
        response_method: Union[str, list[str]],
        *,
        metric_name: Optional[str] = None,
        data_source: Literal["test", "train", "X_y"] = "test",
        X: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> pd.DataFrame: ...



"""Custom help formatter for the CLI."""

import re
import shutil
from argparse import ArgumentParser, HelpFormatter

from rich.console import Console
from rich.theme import Theme

skore_console_theme = Theme(
    {
        "repr.str": "cyan",
        "rule.line": "orange1",
        "repr.url": "orange1",
    }
)


class RichColorHelpFormatter(HelpFormatter):
    """Custom help formatter for the CLI."""

    def __init__(self, prog, indent_increment=2, max_help_position=24, width=None):
        width = shutil.get_terminal_size()[0] if width is None else width
        super().__init__(prog, indent_increment, max_help_position, width)
        self.console = Console(theme=skore_console_theme)

    def _format_action_invocation(self, action):
        """Format the action invocation (flags and arguments)."""
        if not action.option_strings:
            metavar = self._metavar_formatter(action, action.dest)(1)[0]
            return metavar
        else:
            parts = []
            # Format short options
            if action.option_strings:
                parts.extend(
                    f"[cyan bold]{opt}[/cyan bold]" for opt in action.option_strings
                )
            # Format argument
            if action.nargs != 0:
                default = self._get_default_metavar_for_optional(action)
                args_string = self._format_args(action, default)
                parts.append(f"[orange1 bold]{args_string}[/orange1 bold]")

            return " ".join(parts)

    def _format_usage(self, usage, actions, groups, prefix):
        """Format the usage line."""
        if prefix is None:
            prefix = "usage: "

        # Format the usage line
        formatted = super()._format_usage(usage, actions, groups, prefix)

        # Apply rich formatting
        formatted = re.sub(r"usage:", "[orange1 bold]usage:[/orange1 bold]", formatted)
        formatted = re.sub(
            r"(?<=\[)[A-Z_]+(?=\])", lambda m: f"[cyan]{m.group()}[/cyan]", formatted
        )

        return formatted

    def format_help(self):
        """Format the help message."""
        help_text = super().format_help()

        # Format section headers
        help_text = re.sub(
            r"^([a-zA-Z ]+ arguments:)$",
            r"[dim]\1[/dim]",
            help_text,
            flags=re.MULTILINE,
        )

        # Format default values
        help_text = re.sub(
            r"\(default: .*?\)", lambda m: f"[dim]{m.group()}[/dim]", help_text
        )

        # Color the subcommands in cyan
        help_text = re.sub(
            r"(?<=\s)(launch|create|quickstart)(?=\s+)",
            r"[cyan bold]\1[/cyan bold]",
            help_text,
        )

        # Color "options" in orange1
        help_text = re.sub(
            r"(?<=\s)(options)(?=:)",
            r"[orange1]\1[/orange1]",
            help_text,
        )

        return help_text


class ColorArgumentParser(ArgumentParser):
    """Custom argument parser for the CLI."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, formatter_class=RichColorHelpFormatter)

    def print_help(self, file=None):
        """Print the help message."""
        console = Console(file=file)
        console.print(self.format_help())



"""Base class for all items in the project."""

from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional
from uuid import uuid4

from jinja2 import Environment, FileSystemLoader


class ItemTypeError(Exception):
    """Item type exception.

    Exception raised when an attempt is made to convert an object to an Item, but the
    object's type is not supported.
    """


class Item(ABC):
    """
    Abstract base class for all items in the project.

    This class provides a common interface for all items, including
    creation and update timestamps.

    Parameters
    ----------
    created_at : Union[str, None], optional
        The creation timestamp of the item. If None, the current time is used.
    updated_at : Union[str, None], optional
        The last update timestamp of the item. If None, the current time is used.
    note : Union[str, None]
        An optional note.

    Attributes
    ----------
    created_at : str
        The creation timestamp of the item.
    updated_at : str
        The last update timestamp of the item.
    note : Union[str, None]
        An optional note.
    """

    def __init__(
        self,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        note: Optional[str] = None,
    ):
        now = datetime.now(tz=timezone.utc).isoformat()

        self.created_at = created_at or now
        self.updated_at = updated_at or now
        self.note = note

    @classmethod
    @abstractmethod
    def factory(cls, *args, **kwargs) -> Item:
        """
        Create and return a new instance of the Item.

        Returns
        -------
        Item
            A new instance of the Item.
        """

    @property
    def __parameters__(self) -> dict[str, Any]:
        """
        Get the parameters of the Item instance.

        Returns
        -------
        dict[str, Any]
            A dictionary containing the parameters of the Item instance.
        """
        cls = self.__class__
        cls_parameters = inspect.signature(cls).parameters

        return {parameter: getattr(self, parameter) for parameter in cls_parameters}

    def __repr__(self) -> str:
        """Represent the item."""
        return f"{self.__class__.__name__}(...)"

    def _repr_mimebundle_(self, include=None, exclude=None):
        return {"text/html": self._repr_html_()}

    def _repr_html_(self):
        """Represent the item in a notebook."""
        item_folder = Path(__file__).resolve().parent
        templates_env = Environment(loader=FileSystemLoader(item_folder))
        template = templates_env.get_template("standalone_widget.html.jinja")

        static_files_path = item_folder.parent / "ui" / "static" / "assets"

        def read_asset_content(path):
            with open(static_files_path / path) as f:
                return f.read()

        script_content = read_asset_content("index.js")
        styles_content = read_asset_content("index.css")

        context = {
            "id": uuid4().hex,
            "item": self.as_serializable_dict(),
            "script": script_content,
            "styles": styles_content,
        }

        return template.render(**context)

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend.

        Derived class must call their super implementation and merge the result with
        their output.
        """
        return {
            "updated_at": self.updated_at,
            "created_at": self.created_at,
            "note": self.note,
        }



"""MediaItem.

This module defines the MediaItem class, used to persist media.
"""

from __future__ import annotations

from enum import Enum
from typing import Any, Optional

from .item import Item, ItemTypeError


def lazy_is_instance(object: Any, cls_fullname: str) -> bool:
    """Return True if object is an instance of `cls_fullname`."""
    return cls_fullname in {
        f"{cls.__module__}.{cls.__name__}" for cls in object.__class__.__mro__
    }


class MediaType(Enum):
    """A media type of a string."""

    HTML = "text/html"
    MARKDOWN = "text/markdown"
    SVG = "image/svg+xml"


class MediaItem(Item):
    """A class used to persist media."""

    def __init__(
        self,
        media: str,
        media_type: str,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        note: Optional[str] = None,
    ):
        """
        Initialize a MediaItem.

        Parameters
        ----------
        media : str
            The media content.
        media_type : str
            The MIME type of the media content.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        note : str, optional
            A note.
        """
        super().__init__(created_at, updated_at, note)

        self.media = media
        self.media_type = media_type

    @classmethod
    def factory(
        cls,
        media: str,
        /,
        media_type: Optional[str] = MediaType.MARKDOWN.value,
        **kwargs,
    ) -> MediaItem:
        """
        Create a new MediaItem instance from a string.

        Parameters
        ----------
        media : str
            The media content.
        media_type : str
            The MIME type of the media content.

        Returns
        -------
        MediaItem
            A new MediaItem instance.
        """
        if not isinstance(media, str):
            raise ItemTypeError(f"Type '{media.__class__}' is not supported.")

        # | Before Python 3.12, a TypeError is raised if a non-Enum-member is used in a
        # | containment check.
        #
        # https://docs.python.org/3.12/library/enum.html#enum.EnumType.__contains__
        if media_type not in MediaType._value2member_map_:
            raise ValueError(f"MIME type '{media_type}' is not supported.")

        return cls(media, media_type, **kwargs)

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        return super().as_serializable_dict() | {
            "media_type": self.media_type,
            "value": self.media,
        }



"""'High class imbalance (too few examples)' warning.

This warning is shown when some class in a dataset has too few examples in the test set.
"""

from __future__ import annotations

from collections import Counter
from typing import TYPE_CHECKING, Any, Optional, Union

from skore.sklearn.train_test_split.warning.train_test_split_warning import (
    TrainTestSplitWarning,
)

if TYPE_CHECKING:
    from skore.sklearn.cross_validate import MLTask

    ArrayLike = Any


class HighClassImbalanceTooFewExamplesWarning(TrainTestSplitWarning):
    """Check whether the test set has too few examples in one class."""

    MSG = (
        "It seems that you have a classification problem with at least one class with "
        "fewer than 100 examples in the test set. In this case, using train_test_split "
        "may not be a good idea because of high variability in the scores obtained on "
        "the test set. We suggest three options to tackle this challenge: you can "
        "increase test_size, collect more data, or use skore's cross_validate function."
    )

    @staticmethod
    def check(
        y_test: Optional[ArrayLike],
        stratify: Optional[ArrayLike],
        y_labels: Optional[ArrayLike],
        ml_task: MLTask,
        **kwargs,
    ) -> Union[str, None]:
        """Check whether the test set has too few examples in one class.

        More precisely, we check whether some class in `y_labels` has
        fewer than 100 examples in `y_test`.
        The other arguments are needed to see if the check is relevant. For
        example, if `y_test` is used for a regression task, then the check should
        be skipped.

        Parameters
        ----------
        y_test : array-like or None
            A 1-dimensional target vector, as a list, numpy array, scipy sparse array,
            or pandas dataframe.
        stratify : array-like or None
            An 1-dimensional target array to be used for stratification.
        y_labels: array-like or None
            A 1-dimensional array containing the class labels
            (e.g. [0, 1] if the task is binary classification),
            if the task is classification.
            This is needed in the case where `y_test` contains zero examples
            for some class.
        ml_task : MLTask
            The type of machine-learning tasks being performed.

        Returns
        -------
        warning
            None if the check passed, otherwise the warning message.
        """
        if (
            stratify
            or (y_test is None or len(y_test) == 0)
            or (y_labels is None or len(y_labels) == 0)
            or ("classification" not in ml_task)
        ):
            return None

        counts = Counter(y_test)

        if all(counts.get(class_, 0) >= 100 for class_ in y_labels):
            return None

        return HighClassImbalanceTooFewExamplesWarning.MSG



"""Types between parts of the sklearn module."""

from typing import Literal

MLTask = Literal[
    "binary-classification",
    "multiclass-classification",
    "regression",
    "clustering",
    "unsupported",
]



from sklearn.pipeline import Pipeline


def _check_supported_estimator(supported_estimators):
    def check(accessor):
        estimator = accessor._parent.estimator_
        if isinstance(estimator, Pipeline):
            estimator = estimator.steps[-1][1]
        supported_estimator = isinstance(estimator, supported_estimators)

        if not supported_estimator:
            raise AttributeError(
                f"The {estimator.__class__.__name__} estimator is not supported "
                "by the function called."
            )

        return True

    return check



"""PolarsDataFrameItem.

This module defines the PolarsDataFrameItem class,
which represents a polars DataFrame item.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from .item import Item, ItemTypeError

if TYPE_CHECKING:
    import polars


class PolarsToJSONError(Exception):
    """Something happened while converting a polars DataFrame to JSON."""


class PolarsDataFrameItem(Item):
    """
    A class to represent a polars DataFrame item.

    This class encapsulates a polars DataFrame along with its
    creation and update timestamps.
    """

    def __init__(
        self,
        dataframe_json: str,
        created_at: Union[str, None] = None,
        updated_at: Union[str, None] = None,
        note: Union[str, None] = None,
    ):
        """
        Initialize a PolarsDataFrameItem.

        Parameters
        ----------
        dataframe_json : str
            The JSON representation of the dataframe, without its index.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        note : Union[str, None]
            An optional note.
        """
        super().__init__(created_at, updated_at, note)

        self.dataframe_json = dataframe_json

    @property
    def dataframe(self) -> polars.DataFrame:
        """
        The polars DataFrame from the persistence.

        Its content can differ from the original dataframe because it has been
        serialized using polars' `to_json` function and not pickled, in order to be
        environment-independent.
        """
        import io

        import polars

        with io.StringIO(self.dataframe_json) as df_stream:
            dataframe = polars.read_json(df_stream)
            return dataframe

    @classmethod
    def factory(cls, dataframe: polars.DataFrame, /, **kwargs) -> PolarsDataFrameItem:
        """
        Create a new PolarsDataFrameItem instance from a polars DataFrame.

        Parameters
        ----------
        dataframe : polars.DataFrame
            The polars DataFrame to store.

        Returns
        -------
        PolarsDataFrameItem
            A new PolarsDataFrameItem instance.

        Notes
        -----
        The dataframe must be JSON serializable.
        """
        import polars

        if not isinstance(dataframe, polars.DataFrame):
            raise ItemTypeError(f"Type '{dataframe.__class__}' is not supported.")

        try:
            dataframe_json = dataframe.write_json()
        except Exception as e:
            raise PolarsToJSONError("Conversion to JSON failed") from e

        return cls(dataframe_json=dataframe_json, **kwargs)

    def as_serializable_dict(self):
        """Convert item to a JSON-serializable dict to used by frontend."""
        return super().as_serializable_dict() | {
            "value": self.dataframe.to_pandas().fillna("NaN").to_dict(orient="tight"),
            "media_type": "application/vnd.dataframe",
        }
